from Scarf.Base.ThreadPool import ThreadPool

import queue, time, threading


class TaskParamter:
    def __init__(self, interval, times, callback, _this=None, user_arg=None):
        self.interval = interval
        self.cycle = interval
        self.times = times
        self.__callback = callback
        self.__user_arg = user_arg
        self.__this = _this
        self._last_time = 0

    def __call__(self, *args, **kwargs):
        return self.__callback(self.__this, *args, **kwargs)

    def call(self, *args, **kwargs):
        return self.__callback(self.__this, *args, self.__user_arg)


class TaskItem:
    def __init__(self, callback, arg, thread_pool: ThreadPool = None):
        self.callback = callback
        self.arg = arg
        self.__thread_pool = thread_pool

    def __call__(self, *args):
        if self.__thread_pool:
            self.__thread_pool.add_task(self.callback, tuple([self.arg] + [item for item in args]))
        else:
            self.callback(self.arg, *args)


class Timer:
    def __init__(self, _init_time, thread_pool: ThreadPool = None):
        self.__work_queue = [queue.Queue() for item in range(0, 0x3e9)]
        self.__wait_queue = []
        self.__last_time = _init_time
        self.__last_index = 0
        self.__time_instance = 0
        self.__instantaneous_time = 0
        self.__thread_pool = thread_pool
        self.__lock = threading.Lock()

    def __reslove(self, task_item, _time):
        flag = task_item.call(_time)
        if task_item.times > 1 and flag:
            task_item.times -= 1
        elif task_item.times == 1 or not flag:
            return
        task_item.interval += task_item.cycle
        self.add_task(task_item)

    def add_task(self, item: TaskParamter):
        if item.interval < 0x3e9:
            self.__work_queue[item.interval].put_nowait(TaskItem(self.__reslove, item))
        else:
            self.__wait_queue.append(item)

    def __update_wait(self, instance):
        for item in tuple(self.__wait_queue):
            item.interval -= instance
            if item.interval < 0x3e9:
                self.add_task(item)
                self.__wait_queue.remove(item)

    def __emit_queue(self, start, end, *args):
        """
        emit queue start to end
        """
        for _idx in range(start, end + 1):
            if not self.__work_queue[_idx].empty():
                self.__work_queue[_idx].get_nowait()(*args)

    def update(self, _time):
        _instance = _time - self.__last_time
        if _instance > 0x3e8:
            while _instance > 0x3e8:
                self.__emit_queue(self.__last_index, 0x3e8, _time)
                self.__emit_queue(0, self.__last_index, _time)
                self.__last_time = _time
                self.__update_wait(0x3e8)
                _instance -= 0x3e8
            start = 0
            end = _instance
        else:
            start = self.__last_index
            end = _instance

        self.__emit_queue(start, end, _time)
        self.__last_index = end



from Scarf.paramter import Paramter
from Scarf.Base.ThreadPool import ThreadPool
from Scarf.Protocol.ClientRequest import ClientRequest
from Scarf.Context.RequestReslove import RequestReslove
from Scarf.Context.Timer import Timer
from Scarf.Context.Log import LogDeeper

import traceback, sys, asyncio, time


class ClientContext:
    def __init__(self, config: Paramter, routers, timer_tasks, filters, sql, extend, loggers,
                 main_logger):
        self.__config = config
        self.__filters = filters
        self.__event_loop = None
        self.__thread_pool = None
        self.__timer = None
        self.__main_logger = main_logger
        self.__http2 = self.__config.get_http2()
        self.__request_reslove = RequestReslove(routers, self.__error_cache, self.__config, self.__filters, sql, extend,
                                                loggers)
        thread_opts = self.__config.get_thread_pool_worker_num()
        if not thread_opts.get("single"):
            self.__thread_pool = ThreadPool(thread_opts["min"], thread_opts["max"], thread_opts["max_work_task"],
                                            self.__error_cache)
            if len(timer_tasks) > 0:
                self.__timer = Timer(int(time.time() * 1000), self.__thread_pool)
                for item in timer_tasks:
                    self.__timer.add_task(item)
                self.__thread_pool.add_task(self.__update_timer, ())
            self.__thread_pool.begin_work()

    def stop_work(self):
        self.__thread_pool.finish_work()

    def __update_timer(self):
        self.__timer.update(int(time.time() * 1000))
        self.__thread_pool.add_task(self.__update_timer, ())

    def set_loop(self, loop):
        self.__event_loop = loop

    def client_context(self, reader, writer):
        crt = ClientRequest(self, reader, writer, self.__event_loop, self.__config, self.__request_reslove,
                            self.__http2,
                            self.__error_cache)
        asyncio.run_coroutine_threadsafe(self.post_read(crt, reader, crt.alive_timeout), loop=self.__event_loop)

    async def post_read(self, crt, reader, timeout):
        try:
            data = (await reader.read(65536)) if timeout == -1 else (
                await asyncio.wait_for(reader.read(65536), loop=self.__event_loop, timeout=timeout))
            if len(data) == 0:
                raise asyncio.TimeoutError
            if self.__thread_pool:
                self.__thread_pool.add_task(crt.read_data, (data,))
            else:
                crt.read_data(data)
        except BrokenPipeError:
            crt.writer.close()
        except asyncio.exceptions.IncompleteReadError:
            crt.writer.close()
        except asyncio.TimeoutError:
            crt.writer.close()
        except ConnectionResetError:
            crt.writer.close()
        except Exception as e:
            self.__error_cache(e)

    def async_taker(self, method, *args):
        self.__thread_pool.add_task(method, args)

    def __error_cache(self, ex):
        if self.__main_logger:
            self.__main_logger.error(str.join("", traceback.format_exception(ex, value=ex, tb=ex.__traceback__)))
        else:
            traceback.print_exception(ex, value=ex, tb=ex.__traceback__)



import asyncio, ssl, sys


class TCPServer:
    def __init__(self, ports, callback, http2, ssloptions={}):
        (http_port, https_port) = ports
        self.__http_port = http_port
        self.__https_port = https_port
        self.__http2 = http2
        self.__ssl_options = ssloptions
        self.__ssl_context = None
        self.__HTTP_Server = None
        self.__HTTPS_Server = None
        self.__callback = callback
        self.__event_loop = self.__GetEventLoop()
        self.__event_loop.run_until_complete(self.__CreateTcpServer())

    def start_loop(self):
        if self.__HTTPS_Server:
            self.__event_loop.run_until_complete(self.__HTTPS_Server.start_serving())
        if self.__HTTP_Server:
            self.__event_loop.run_until_complete(self.__HTTP_Server.start_serving())
        while True:
            try:
                self.__event_loop.run_forever()
            except asyncio.InvalidStateError as e:
                continue

    def get_event_loop(self):
        return self.__event_loop

    def stop_loop(self):
        self.__event_loop.stop()

    async def __CreateTcpServer(self):
        if self.__http_port > -1:
            self.__HTTP_Server = await asyncio.start_server(self.__callback, start_serving=False, host='0.0.0.0',
                                                            port=self.__http_port, loop=self.__event_loop)
        if self.__https_port > -1:
            self.__ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            if self.__http2 and ssl.HAS_ALPN:
                self.__ssl_context.set_alpn_protocols(["h2"])
            self.__ssl_context.load_cert_chain(self.__ssl_options["crt"], self.__ssl_options["key"],
                                               self.__ssl_options["pwd"])
            self.__HTTPS_Server = await asyncio.start_server(self.__callback, start_serving=False, host='0.0.0.0',
                                                             port=self.__https_port, loop=self.__event_loop,
                                                             ssl_handshake_timeout=self.__ssl_options[
                                                                 "ssl_handshake_timeout"],
                                                             ssl=self.__ssl_context)

    def __GetEventLoop(self):
        """Get The Best Event Loop At This Platform"""
        loop = None
        if sys.platform == 'win32':
            loop = asyncio.ProactorEventLoop()
        else:
            loop = asyncio.SelectorEventLoop(selector=None)
        return loop



import threading, types, queue


class Actuator:
    """
    Thread Reslove Object
    """

    def __init__(self, function: types.FunctionType, arguments: tuple):
        self.arguments = arguments
        self.function = function

    def __call__(self, *args):
        self.function(*args)


class ThreadContext:
    def __init__(self, _id: int, revicer_: types.FunctionType, _queue: queue.Queue, error_: types.FunctionType):
        self.sync_flag = True
        self.__id = _id
        """ Sync Work Support """
        self.__revicer = revicer_
        """ Async Work Support """
        self.__queue = _queue
        self.__event = threading.Event()
        self.__error = error_
        self.__th = threading.Thread(target=self.__revicer, args=(_id,))
        self.__finish_flag = False
        self.__event.set()

    def reslove_context(self, destory_notify):
        timeout = -1 if destory_notify is None else 0.5
        alive_times = 6
        while not self.__finish_flag:
            self.__event.wait()
            if self.sync_flag:
                context = yield None
                if context is None:
                    continue
                result = context.function(*context.arguments)
                yield result
            try:
                context = self.__queue.get(block=timeout > -1, timeout=timeout)
                context(*context.arguments)
            except queue.Empty:
                if destory_notify is None:
                    self.__event.clear()
                    continue
                else:
                    alive_times -= 1
                    if alive_times == 0:
                        destory_notify(self.__id)
                        return
            except Exception as e:
                self.__error(e)

    def start_work(self):
        self.__th.start()

    def destroy(self):
        self.__finish_flag = True
        self.active()

    def active(self):
        self.__event.set()

    def sleep(self):
        self.__event.clear()


class ThreadPool:
    def __init__(self, min_size: int, max_size: int, max_work_num: int, error_: types.FunctionType):
        """
        Thread Pool Constructor
        :param min_size: Thread Pool Max Thread Size
        :param max_size: Thread Pool Min Thread Size
        :param error_: Thread Pool Reslove Error Callback
        """
        self.__min_size = int(min_size)
        self.__max_size = int(max_size)
        self.__max_work_num = int(max_work_num)
        self.__error_callback = error_
        self.__all_queue_work = False
        self.__operator_lock = threading.Lock()
        self.__recivers = [_id for _id in range(0, max_size)]
        self.__Queues = [{'source': queue.Queue(), 'id': _} for _ in range(0, min_size)]
        self.__threads = [
            (ThreadContext(_id, self.__get_revicer, self.__Queues[_id]["source"], error_) if _id < min_size else None)
            for _id in
            range(0, max_size)]

    def __get_revicer(self, _id: int):
        try:
            self.__recivers[_id] = self.__threads[_id].reslove_context(
                None if _id < self.__min_size else self.__destory_notify)
            self.__threads[_id].sync_flag = False
            self.__recivers[_id].send(None)
            self.__threads[_id].sleep()
        except StopIteration:
            return
        except Exception as e:
            print(str(e))

    def __comput_better(self):
        _queue = self.__Queues.copy()
        _queue.sort(key=lambda el: el["source"].qsize())
        return _queue[0]["id"]

    def __destory_notify(self,_id):
        self.__operator_lock.acquire()
        self.__Queues.remove(self.__Queues[_id])
        self.__threads[_id].destroy()
        self.__threads.remove(self.__threads[_id])
        self.__all_queue_work = False
        self.__operator_lock.release()

    def add_task(self, fn, args):
        index = self.__comput_better()
        if self.__Queues[index]["source"].qsize() >= self.__max_work_num and not self.__all_queue_work:
            self.__operator_lock.acquire()
            for _id, _th in enumerate(self.__threads):
                if _th is None:
                    self.__Queues.append({'source': queue.Queue(), 'id': _id})
                    self.__threads[_id] = ThreadContext(_id, self.__get_revicer, self.__Queues[_id]["source"],
                                                        self.__error_callback)
                    self.__threads[_id].start_work()
                    if _id == self.__max_size - 1:
                        self.__all_queue_work = True
                    break
            self.__operator_lock.release()
        self.__Queues[index]['source'].put(Actuator(fn, args))
        self.__threads[index].active()

    def add_task_sync_reslove(self, fn: types.MethodType, args: tuple):
        index = self.__comput_better()
        self.__threads[index].sync_flag = True
        self.__threads[index].active()
        result = self.__recivers[index].send(Actuator(fn, args))
        self.__recivers[index].send(None)
        self.__threads[index].sync_flag = False
        return result

    def begin_work(self):
        for _th in self.__threads:
            if _th is None:
                continue
            _th.start_work()

    def finish_work(self):
        for _th in self.__threads:
            if _th is None:
                continue
            _th.destroy()


from string import Template
import logging, sys, os, traceback

if sys.platform == "win32":
    import ctypes, sys

    STD_INPUT_HANDLE = -10
    STD_OUTPUT_HANDLE = -11
    STD_ERROR_HANDLE = -12

    FOREGROUND_BLACK = 0x00  # black.
    FOREGROUND_DARKBLUE = 0x01  # dark blue.
    FOREGROUND_DARKGREEN = 0x02  # dark green.
    FOREGROUND_DARKSKYBLUE = 0x03  # dark skyblue.
    FOREGROUND_DARKRED = 0x04  # dark red.
    FOREGROUND_DARKPINK = 0x05  # dark pink.
    FOREGROUND_DARKYELLOW = 0x06  # dark yellow.
    FOREGROUND_DARKWHITE = 0x07  # dark white.
    FOREGROUND_DARKGRAY = 0x08  # dark gray.
    FOREGROUND_BLUE = 0x09  # blue.
    FOREGROUND_GREEN = 0x0a  # green.
    FOREGROUND_SKYBLUE = 0x0b  # skyblue.
    FOREGROUND_RED = 0x0c  # red.
    FOREGROUND_PINK = 0x0d  # pink.
    FOREGROUND_YELLOW = 0x0e  # yellow.
    FOREGROUND_WHITE = 0x0f  # white.

    # Windows CMD命令行 背景颜色定义 background colors
    BACKGROUND_BLUE = 0x10  # dark blue.
    BACKGROUND_GREEN = 0x20  # dark green.
    BACKGROUND_DARKSKYBLUE = 0x30  # dark skyblue.
    BACKGROUND_DARKRED = 0x40  # dark red.
    BACKGROUND_DARKPINK = 0x50  # dark pink.
    BACKGROUND_DARKYELLOW = 0x60  # dark yellow.
    BACKGROUND_DARKWHITE = 0x70  # dark white.
    BACKGROUND_DARKGRAY = 0x80  # dark gray.
    BACKGROUND_BLUE = 0x90  # blue.
    BACKGROUND_GREEN = 0xa0  # green.
    BACKGROUND_SKYBLUE = 0xb0  # skyblue.
    BACKGROUND_RED = 0xc0  # red.
    BACKGROUND_PINK = 0xd0  # pink.
    BACKGROUND_YELLOW = 0xe0  # yellow.
    BACKGROUND_WHITE = 0xf0  # white.

    # get handle
    std_out_handle = ctypes.windll.kernel32.GetStdHandle(STD_OUTPUT_HANDLE)


    def set_cmd_text_color(color, handle=std_out_handle):
        Bool = ctypes.windll.kernel32.SetConsoleTextAttribute(handle, color)
        return Bool


    def printWhite(mess):
        set_cmd_text_color(FOREGROUND_WHITE)
        sys.stdout.write(mess)
        resetColor()


    def printSkyBlue(mess):
        set_cmd_text_color(FOREGROUND_SKYBLUE)
        sys.stdout.write(mess)
        resetColor()


    def printRed(mess):
        set_cmd_text_color(FOREGROUND_RED)
        sys.stdout.write(mess)
        resetColor()


    def printYellow(mess):
        set_cmd_text_color(FOREGROUND_YELLOW)
        sys.stdout.write(mess)
        resetColor()


    def printPink(mess):
        set_cmd_text_color(FOREGROUND_PINK)
        sys.stdout.write(mess)
        resetColor()


    # reset white
    def resetColor():
        set_cmd_text_color(FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE)


class LogDeeper(logging.Handler):
    def __init__(self, filename, module_name, released,
                 format=f"%(asctime)s - $filename[line:%(lineno)d | %(funcName)s] - %(levelname)s: %(message)s"):
        super().__init__()
        if not os.path.exists("./logs"):
            os.mkdir("./logs")
        if not os.path.exists("./logs/" + module_name + ".log"):
            open("./logs/" + module_name + ".log", "w+").close()
        self.__released = released
        self.__module_name = module_name
        self.__file = filename
        self.__logger_handle = []
        self.__logger = logging.getLogger(module_name)
        self.__logger.setLevel(logging.DEBUG)
        self.__file_log_handle = logging.FileHandler("./logs/" + module_name + ".log", mode="a+", encoding='utf8')
        self.__file_log_handle.setLevel(logging.NOTSET)
        format = Template(format).substitute(filename=filename)
        self.__file_log_handle.setFormatter(logging.Formatter(format))
        self.__logger.addHandler(self.__file_log_handle)
        self.__logger.addHandler(self)
        self.warn = self.__logger.warning
        self.critical = self.__logger.critical
        self.error = self.__logger.error
        if self.__released:
            self.__logger.debug = lambda _: _
        else:
            self.debug = self.__logger.debug
        self.info = self.__logger.info

    def handle(self, record):
        if sys.platform == "win32":
            print_func = None
            if record.levelname == "INFO":
                print_func = printWhite
            elif record.levelname == "DEBUG":
                print_func = printSkyBlue
            elif record.levelname == "WARNING":
                print_func = printYellow
            elif record.levelname == "ERROR":
                print_func = printRed
            elif record.levelname == "CRITICAL":
                print_func = printPink
            else:
                print_func = print
            print_func(record.msg+"\n")
        else:
            levelColor = 37
            if record.levelname == "INFO":
                levelColor = 37
            elif record.levelname == "DEBUG":
                levelColor = 44
            elif record.levelname == "WARNING":
                levelColor = 33
            elif record.levelname == "ERROR":
                levelColor = 31
            elif record.levelname == "CRITICAL":
                levelColor = 35
            print("\033[%im %s \033[0m" % (levelColor, record.msg,))
        for item in self.__logger_handle:
            item(record)

    def add_handle(self, handle):
        self.__logger_handle.append(handle)



from Scarf.Context.RequestReslove import RouterRegister, FilterRegister, EventHandler, WebSocketParamter
from Scarf.paramter import Paramter
from Scarf.Base.TCPServer import TCPServer
from Scarf.Context.Timer import TaskParamter
from Scarf.Context.Context import ClientContext
from Scarf.Protocol.HTTP import HTTPRequest
from Scarf.Tip import *
from Scarf.Context.Log import LogDeeper

import types, os, yaml, json, sys


class Scarf:
    def __init__(self):
        """transport server"""
        self.__transport_server = None
        self.__config: Paramter = None
        self.__context: ClientContext = None
        self.__routers = []
        self.__req_intercept = []
        self.__res_filter = []
        self.__events_ = []
        self.__tasks = []
        self.__sql_opts = {}
        self.__websocket_map = {}
        self.__extend = {}
        self.__module_config = {}
        self.__modules = {}
        self.__loggers = {}
        self.__server_log: LogDeeper = None

    def __scan_config(self):
        if os.path.exists("./config"):
            for item in os.walk("./config"):
                for file in item[2]:
                    module, exter_name = os.path.splitext(file)
                    with open("./config/" + file, "r+", encoding='utf8') as fd:
                        if exter_name == ".yml":
                            self.__module_config[module] = yaml.safe_load(fd.read())
                        elif exter_name == ".json":
                            self.__module_config[module] = json.loads(fd.read())
                        else:
                            self.__server_log.warn("ignore '%s' module config ,file type not support" % (module,))
                        fd.close()

    def load_config_from_file(self, config_path: str):
        self.__config = Paramter(config_path)
        self.__server_log = LogDeeper("Scarf", "WebServer", self.__config.get_release())

    def load_config(self, config: dict):
        self.__config = Paramter(None, config)
        self.__server_log = LogDeeper("Scarf", "WebServer", self.__config.get_release())

    def scan_module(self, md):
        for item in dir(md):
            value = getattr(md, item)
            if isinstance(value, types.MethodType):
                ret_type = value.__annotations__.get("return")
                if ret_type is RouterRegister:
                    self.__routers.append(value(item))
                elif ret_type is TaskParamter:
                    self.__tasks.append(value(item))
                elif ret_type is WebSocketParamter:
                    value = value(item)
                    if not self.__websocket_map.get(value.path):
                        self.__websocket_map[value.path] = []
                    self.__websocket_map[value.path].append(value)
            self.__modules[md.__module__ + "." + md.__class__.__name__] = md

        if not hasattr(md, '_emit'):
            setattr(md, '_emit', self.__emit)

    def __emit(self, event_, *params, **kwargs):
        sources = []
        event_obj = None
        for i, e in enumerate(self.__events_):
            result = e.test_event(event_)
            if isinstance(result, bool):
                if result:
                    event_obj = e
                    break
            elif isinstance(result, int) and result > 0:
                sources.append([result, i])

        if len(sources) > 0 and event_obj is None:
            sources.sort(key=lambda a: a[0], reverse=True)
            event_obj = self.__events_[sources[0][1]]

        if event_obj is None:
            return None

        return event_obj(*params, **kwargs)

    def register_hook(self, module):
        if isinstance(module, FilterRegister):
            _req_flag, _res_flag = module.get_reg_opt()
            if _req_flag:
                self.__req_intercept.append(module)
            if _res_flag:
                self.__res_filter.append(module)
            setattr(module, '_emit', self.__emit)
        else:
            raise TypeError("Class Not 'FilterRegister' ")

    def register_events(self, md):
        for item in dir(md):
            value = getattr(md, item)
            if isinstance(value, types.MethodType):
                ret_type = value.__annotations__.get("return")
                if ret_type is EventHandler:
                    self.__events_.append(value(item))
        if not hasattr(md, '_emit'):
            setattr(md, '_emit', self.__emit)

    def register_global(self, name, module):
        self.__extend[name] = module

    def register_sql_model(self, name, model):
        for item in self.__config.get_sql_options():
            if name == str(item["name"]):
                item = item.copy()
                del item["name"]
                value = model(**item)
                if isinstance(value, SQLModel):
                    self.__sql_opts[name] = value
                    return
                else:
                    raise ValueError("Invaild SQL Factory Class")

        raise ValueError("No The Same DataSource ('%s')" % (name,))

    def __router_map(self, router, path, static_path, entry_file):
        router = str.join('', [item for item in router.split("/*") if len(item.strip()) > 0])

        def static_factory(_this, req: HTTPRequest):
            file_path = req.path[req.path.index(router) + len(router):]
            not_found_404 = (404, {"Content-Type": "text/plain"}, "Not Found")
            if len(file_path.strip()) == 0:
                file_path = static_path + path + "/" + entry_file
                if os.path.exists(file_path):
                    return FileDeliver(file_path)
                else:
                    return not_found_404
            else:
                file_path = static_path + path + file_path
                if not os.path.exists(file_path):
                    req.is_not_found = True
                    return not_found_404
                elif os.path.isfile(file_path):
                    return FileDeliver(file_path)
                else:
                    file_path += ('/' + entry_file)
                    if os.path.exists(file_path):
                        return FileDeliver(file_path)
                    else:
                        return not_found_404

        return static_factory

    def start_server(self):
        if not self.__config:
            self.__config = Paramter(configpath=None, config={})
        self.__scan_config()
        static_path, entry_file = self.__config.get_static_path()
        static_path = static_path.replace("\\", "/")
        for page in self.__config.get_pages():
            path = FileDeliver(page["path"]).test_exist(static_path)
            result = None
            if path:
                if os.path.isfile(path):
                    result = lambda _: FileDeliver(page["path"])
                else:
                    result = self.__router_map(page["router"], page["path"], static_path, entry_file)
            self.__routers.append(
                RouterRegister(page["router"], (Methods.GET,), ((), (), (), ()), result,
                               None, True))
        for item in self.__modules.items():
            name, module = item
            filename = sys.modules[module.__module__].__file__.replace('/', "\\").split(
                os.path.split(sys.path[0])[0].replace('/', "\\") + "\\")[1]
            module_name = module.__module__ + "." + module.__class__.__name__
            logger = LogDeeper(filename, module_name, self.__config.get_release(),
                               self.__module_config.get(module_name, {}).get("log_format",
                                                                             f"%(asctime)s - $filename[line:%(lineno)d | %(funcName)s] - %(levelname)s: %(message)s"))
            self.__loggers[module_name] = logger

        if len(self.__websocket_map.keys()) > 0:
            handshake_timeout = -1
            data_timeout = -1
            for item in self.__websocket_map.items():
                k, v = item
                hooks = [None, None, None, None]
                for _ in v:
                    if _.hook == WebSocketHOOK.HANDSHAKE:
                        handshake_timeout = _.kwargs.get("handshake_timeout", 10)
                        hooks[0] = _
                    elif _.hook == WebSocketHOOK.MESSAGE:
                        handshake_timeout = _.kwargs.get("data_timeout", 10)
                        hooks[1] = _
                    elif _.hook == WebSocketHOOK.CLOSE:
                        hooks[2] = _
                    elif _.hook == WebSocketHOOK.PING:
                        hooks[3] = _
                try:
                    if hooks.index(None) < 3 and hooks.index(None) > -1:
                        _idx = hooks.index(None)
                        raise Exception("Websocket Hook Not Implementation '%s'" % (
                            'HANDSHAKE' if _idx == 0 else "MESSAGE" if _idx == 1 else "CLOSE" if _idx == 2 else ''))
                except ValueError:
                    pass
                self.__routers.append(
                    WebSocketParamter(k, tuple(hooks), None, handshake_timeout=handshake_timeout,
                                      data_timeout=data_timeout))

        self.__context = ClientContext(self.__config, self.__routers, self.__tasks,
                                       (self.__req_intercept, self.__res_filter),
                                       self.__sql_opts, self.__extend, self.__loggers, self.__server_log)

        for item in self.__module_config.items():
            module_, config = item
            module = self.__modules.get(module_)
            if module:
                name = "init_handle"
                if hasattr(module, config[name]):
                    init_handle = getattr(module, config[name])
                    result = {}
                    for item in init_handle.__annotations__.items():
                        arg_name, tp = item
                        if tp is LogHandle:
                            result[arg_name] = self.__loggers[module.__module__ + "." + module.__class__.__name__]
                        elif isinstance(tp, LogHandle):
                            if tp.name == "WebServer":
                                result[arg_name] = self.__server_log
                            else:
                                result[arg_name] = self.__loggers.get(tp.name)
                        elif tp is SQLModel.DataBaseConnection:
                            if len(self.__sql_opts.values()) > 0:
                                result[arg_name] = tuple(self.__sql_opts.values())[0].get_con()
                            else:
                                result[arg_name] = None
                        elif tp is async_taker:
                            result[arg_name] = self.__context.async_taker
                        elif isinstance(tp, SQLModel.DataBaseConnection):
                            for item in self.__sql_opts.items():
                                name, value = item
                                if name == tp.id:
                                    result[arg_name] = value.get_con()
                                    break
                            else:
                                result[arg_name] = None
                        elif tp is dict or tp is list:
                            result[arg_name] = config.get("config")
                        else:
                            for _ in self.__extend.items():
                                k, v = _
                                if isinstance(tp, str):
                                    if k == tp:
                                        result[arg_name] = v
                                        break
                                elif isinstance(tp, v) or tp is v:
                                    result[arg_name] = tp
                                    break
                    init_handle(**result)
                else:
                    self.__server_log.warn("module '%s' not has init_handle" % (module_,))

        self.__transport_server = TCPServer(self.__config.get_running_time_port(), self.__context.client_context,
                                            self.__config.get_http2()["enable"],
                                            self.__config.get_ssl_options())
        self.__context.set_loop(self.__transport_server.get_event_loop())
        self.__transport_server.start_loop()

    def stop_work(self):
        self.__transport_server.stop_loop()
        self.__context.stop_work()



import os, types, mimetypes


class RequestsState:
    NEXT = 0x0
    PUSHNOW = 0x1
    CLOSE = 0x2


class Methods:
    GET = "GET"
    HEAD = 'HEAD'
    POST = "POST"
    PUT = 'PUT'
    DELETE = 'DELETE'
    CONNECT = 'CONNECT'
    OPTIONS = 'OPTIONS'
    TRACE = "TRACE"


class ClassSource:
    FORM_URLENCODE = 0x01
    JSON = 0x02
    FORM_DATA = 0x04


class FileDeliver:
    @staticmethod
    def read_all(path):
        fd = open(path, mode="rb+")
        _data = fd.read()
        fd.close()
        return _data

    @staticmethod
    def write_all(path, data):
        fd = open(path, mode="wb+")
        _data = fd.write(data)
        fd.close()

    def __init__(self, path=None, data=None, filename=None, filetype=None):
        self.data = None
        self.__path = None
        self.filename = ""
        self.filtype = ""
        if path is None:
            self.data = data
            self.filename = filename
            self.filtype = filetype
            self.ext_name = os.path.splitext(self.filename)[-1] if self.filename else ""
        elif isinstance(path, FileDeliver):
            self.data = path.data
            self.filename = path.filename
            self.filtype = path.filtype
            self.__path = path.__path
            self.ext_name = path.ext_name
        else:
            self.__path = path.replace("\\", "/")
            self.filename = os.path.split(self.__path)[-1]
            _types = mimetypes.guess_type(self.__path)
            self.filtype = _types[0] if len(_types) > 0 else None
            self.ext_name = os.path.splitext(self.__path)[-1]

    def invaild(self):
        return self.__path is None and self.data is None

    def test_exist(self, static_path):
        if os.path.exists(self.__path):
            return self.__path
        elif os.path.exists(static_path + "/" + self.__path):
            return static_path + self.__path
        return None

    def save(self, path):
        FileDeliver.write_all(path, self.data)


class Vector:
    def __init__(self, tp):
        self.type = tp

    def __call__(self):
        return None


class LogHandle:
    def __init__(self, name):
        self.name = name

    def add_handle(self):
        pass

    def info(self):
        pass

    def warn(self):
        pass

    def critical(self):
        pass

    def error(self):
        pass


def async_taker(method, *args):
    pass


class SQLModel:
    @staticmethod
    def model_factory(parent):
        def create_model(cls):
            cache_fields = {}
            __fields__ = []
            for item in cls.__dict__.items():
                key, value = item
                if not ((str(key).startswith("__") and str(key).endswith("__")) or isinstance(value,
                                                                                              types.FunctionType)):
                    __fields__.append(key)
                cache_fields[key] = value
            cache_fields["__fields__"] = __fields__
            return type(cls.__name__, (parent,), cache_fields)

        return create_model

    class DataBaseConnection:
        def __init__(self, id):
            self.id = str(id)

    def get_con(self, *args):
        raise ModuleNotFoundError("Connect Method Not Defined")

    def destory_con(self, *args):
        raise ModuleNotFoundError("Close Method Not Defined")


class WebSocketHOOK:
    HANDSHAKE = 0x0
    MESSAGE = 0x1
    CLOSE = 0x2
    PING = 0x3



from asyncio import StreamWriter, StreamReader
from urllib.parse import unquote
from Scarf.Tip import FileDeliver
import json, types, gzip, asyncio, sys, os, mimetypes, time, hashlib, datetime, traceback


class HTTPConnectionState:
    HEADER = 0x0
    DATA = 0x1
    COMPLATE = 0x2
    INVAILD = 0x3


class HTTPRequest:
    def __init__(self, end_point, stream_id=-1):
        self.method = None
        self.path = None
        self.http_version = None
        self.stream_id = stream_id
        self.__remote_endpoint = end_point
        self.__querys = {}
        self.__headers = {}
        self.cookies = {}
        self.__data = b""
        self.__url_params = None
        self.gzip_flag = False
        self.keep_alive = True
        self.is_file = False
        self.is_websocket = False
        self.is_not_found = False
        self.state = HTTPConnectionState.HEADER
        self.__last_length = 0
        self.__cache_data = b""

    def __parse_cookie(self, cookie_str):
        cookie_str = cookie_str.strip()
        cookies = [item.strip().split('=') for item in cookie_str.split(";") if len(item) > 0]
        for item in cookies:
            self.cookies[item[0].strip()] = item[1].strip()

    def has_complate(self):
        return self.state == HTTPConnectionState.COMPLATE

    def parse_data(self, data):
        if self.state == HTTPConnectionState.HEADER:
            data = self.__cache_data + data
            if data.find(b"\r\n\r\n") == -1:
                self.state = HTTPConnectionState.INVAILD if len(self.__cache_data) > 65536 else self.state
                self.__cache_data += data
            else:
                self.__data = data[data.find(b"\r\n\r\n") + 4:]
                data = data[0:data.find(b"\r\n\r\n")].decode()
                data = data.split("\r\n")
                http_line = data[0].split(' ')
                self.method = http_line[0].upper()
                if http_line[1].find("?") > -1:
                    query_str = http_line[1][http_line[1].find("?") + 1:]
                    for item in query_str.split("&"):
                        name, value = item.split("=")
                        name = unquote(name)
                        value = unquote(value)
                        self.__querys[name] = value
                    self.path = unquote(http_line[1][0:http_line[1].find("?")])
                else:
                    self.path = unquote(http_line[1])
                self.http_version = float(http_line[2][http_line[2].find('/') + 1:])
                if self.http_version == 2.0 and self.stream_id == -1:
                    self.state = HTTPConnectionState.INVAILD
                    return
                for item in data[1:]:
                    item = item.split(':')
                    name = item[0].lower()
                    value = str.join(":", item[1:])
                    if name == 'cookie':
                        self.__parse_cookie(value)
                    self.__headers[name] = value.strip()
                if self.get_header("Connection", "").lower() == "close" or self.get_header(
                        "Connection", "").lower() != "keep-alive" and self.http_version < 1.1:
                    self.keep_alive = False
                encodings = self.get_header("Accept-Encoding", "")
                if encodings:
                    encodings = [item.strip() for item in encodings.split(",")]
                    if "gzip" in encodings:
                        self.gzip_flag = True
                self.__last_length = self.get_content_length() - len(self.__data)
                self.state = HTTPConnectionState.DATA if self.__last_length > 0 else HTTPConnectionState.COMPLATE
                if self.state == HTTPConnectionState.COMPLATE:
                    self.__cache_data = b""
        elif self.state == HTTPConnectionState.DATA:
            self.__data += data
            self.__last_length -= len(data)
            self.state = HTTPConnectionState.DATA if self.__last_length else HTTPConnectionState.COMPLATE

    def set_url_params(self, params):
        self.__url_params = params

    def get_content_length(self):
        return int(self.__headers.get("content-length", 0))

    def del_header(self, name):
        if self.__headers.get(name) is None:
            return
        del self.__headers[name]

    def get_header(self, name, default=None):
        return self.__headers.get(name.lower(), default)

    def set_header(self, name, value):
        self.__headers[name] = str(value)

    def get_data(self):
        return self.__data

    def headers(self):
        return self.__headers.copy()

    def query(self):
        return self.__querys.copy()

    def get_url_params(self):
        return self.__url_params.copy()

    def get_data_with_json(self):
        try:
            return json.loads(self.__data)
        except Exception:
            return {}

    def get_remote_endpoint(self):
        return self.__remote_endpoint._transport._sock.getpeername()

    def get_data_with_form_urlencoded(self):
        params = {}
        try:
            data = self.__data.decode()
            for item in data.split("&"):
                name, value = item.split("=")
                name = unquote(name)
                value = unquote(value)
                params[name] = value
            return params
        except:
            return params

    def get_data_with_form_data(self):
        if self.get_header("content-type", "").strip().lower().find("multipart/form-data") != 0:
            return None
        content_type = self.get_header("content-type", "").strip()
        index = content_type.index("boundary=")
        boundary = content_type[index + 9:]
        if boundary.find('"') > -1 or boundary.find('"') > -1:
            boundary = boundary[1:-1]
        contents = self.__data.split(boundary.encode())[1:-1]

        params = {}
        for item in contents:
            result = [val for val in item.split(b"\r\n") if val != b'--'][1:]
            if len(result) == 3:
                name, value = (result[0], bytes.join(b"\r\n", result[2:]))
            else:
                name, value = (result[0], bytes.join(b"\r\n", result[1:]))
            decode_str = name.decode().split(";")
            n, v = decode_str[1].strip().split("=")
            if n == "name":
                if v.find('"') > -1 and v.find('"') > -1:
                    v = v[1:-1]
                if len(decode_str) > 2:
                    filename_str = decode_str[2].strip().split("=")
                    if filename_str[0] == "filename":
                        file_content_split = value.split(b"\r\n")
                        if file_content_split[0].find(b"Content-Type:") > -1:
                            data = file_content_split[2:]
                            filetype = file_content_split[0].decode().split(":")[1].strip()
                        else:
                            data = file_content_split[1:]
                            filetype = None
                        params[v] = FileDeliver(None, bytes.join(b"\r\n", data),
                                                      filename_str[1][1:-1],
                                                      filetype)
                        continue

                params[v] = value.decode()
        return params


class HTTPResponse:
    static_http_code_plain = {
        "100": "Continue",
        "101": "Switching Protocol",
        "102": "Processing",
        "103": "Early Hints",
        "200": "OK",
        "201": "Created",
        "202": "Accepted",
        "203": "Non-Authoritative Information",
        "204": "No Content",
        "205": "Reset Content",
        "206": "Partial Content",
        "300": "Multiple Choice",
        "301": "Moved Permanently",
        "302": "Found",
        "303": "See Other",
        "304": "Not Modified",
        "307": "Temporary Redirect",
        "308": "Permanent Redirect",
        "400": "Bad Request",
        "401": "Unauthorized",
        "402": "Payment Required",
        "403": "Forbidden",
        "404": "Not Found",
        "405": "Method Not Allowed",
        "406": "Not Acceptable",
        "407": "Proxy Authentication Required",
        "408": "Request Timeout",
        "409": "Conflict",
        "410": "Gone",
        "411": "Length Required",
        "412": "Precondition Failed",
        "413": "Payload Too Large",
        "414": "URI Too Long",
        "415": "Unsupported Media Type",
        "416": "Range Not Satisfiable",
        "417": "Expectation Failed",
        "418": "I'm a teapot",
        "422": "Unprocessable Entity",
        "423": "Locked",
        "424": "Failed Dependency",
        "425": "Too Early",
        "426": "Upgrade Required",
        "428": "Precondition Required",
        "429": "Too Many Requests",
        "431": "Request Header Fields Too Large",
        "451": "Unavailable For Legal Reasons",
        "500": "Internal Server Error",
        "501": "Not Implemented",
        "502": "Bad Gateway",
        "503": "Service Unavailable",
        "505": "HTTP Version Not Supported",
        "506": "Variant Also Negotiates",
        "507": "Insufficient Storage",
        "508": "Loop Detected",
        "510": "Not Extended",
        "511": "Network Authentication Required"
    }

    @staticmethod
    def _guess_type(path):
        return mimetypes.types_map.get(os.path.splitext(path)[-1], "application/octet-stream")

    def __init__(self, loop, writer: StreamWriter, version, config, _gzip):
        mimetypes.types_map[".js"] = "application/javascript"
        self.__loop = loop
        self.__writer = writer
        self.__config = config
        self.__gzip_extern_name = self.__config.get_gzip_file_types()
        self.__static_path = sys.path[0] if len(self.__config.get_static_path()[0]) == 0 else \
            self.__config.get_static_path()[0]
        self.__gzip_flag = _gzip
        self.version = version
        self.__web_cached = self.__config.get_web_cached()
        self.__web_cached["range_size"] -= 1
        self.__code = 200
        self.__plain = "OK"
        self.__headers = {}
        self.__cookies = {}
        self.__start_read = 0
        self.__read_size = 0
        self._source_data = None
        self.__reinit()

    def set_data(self, data):
        self._source_data = data

    def __reinit(self):
        self.__chunked = False
        self.__sender = None
        self.__data = b""
        self.__path = ""
        self.__fd = None
        self.__cache_flag = False
        self.__range_flag = False
        self.__gzip_flag = False
        self.__source_data = None

    def set_code(self, code):
        self.__code = int(code)
        self.__plain = HTTPResponse.static_http_code_plain.get(str(code), "")

    def get_code(self):
        return self.__code

    def get_cookie_by_name(self, name, default_val=None):
        return self.__cookies.get(name, default_val)

    def get_cookie(self):
        cookie_str = []
        for item in self.__cookies.items():
            k, v = item
            if len(v.strip()) == 0:
                cookie_str.append(k)
            else:
                cookie_str.append(("%s=%s" % (str(k), str(v))))
        return str.join('; ', cookie_str)

    def set_cookie(self, val):
        if len(val) == 1:
            self.__cookies[val[0]] = ''
        else:
            k, v = val
            self.__cookies[k] = v

    code = property(get_code, set_code)

    cookies = property(get_cookie, set_cookie)

    def __type_outer(self, result):
        data = b""
        headers = {}
        code = 200
        plain = ""
        if result is None:
            code = 204
            data = b""
        elif isinstance(result, str):
            headers["Content-Type"] = "text/plain"
            data = result.encode()
        elif isinstance(result, dict) or isinstance(result, list):
            headers["Content-Type"] = "application/json"
            for key in (result.keys() if isinstance(result, dict) else range(0, len(result))):
                result[key] = self.__spread_type(result[key])
            data = json.dumps(result).encode()
        elif isinstance(result, tuple):
            for i, item in enumerate(result):
                if i == 0:
                    code = int(item)
                elif i == 1:
                    for _ in dict(result[1]).items():
                        key, value = _
                        headers[key] = value
                elif i == 2:
                    data = bytes(result[2], encoding='utf8') if isinstance(result[2], str) else bytes(result[2])
                elif i == 3:
                    plain = str(result[3])
        elif isinstance(result, types.GeneratorType):
            headers["Transfer-Encoding"] = "chunked"
            data = result
        elif isinstance(result, FileDeliver):
            path = result.test_exist(self.__static_path)
            if path is None:
                code = 404
                headers["Content-Type"] = 'text/plain'
                data = b"Not Found"
            else:
                data = path
        elif isinstance(result, Exception):
            code = 500
            headers["Content-Type"] = "text/plain"
            data = result
        else:
            return self.__type_outer(self.__spread_type(result))
        return (code, plain, headers, data)

    def __spread_type(self, cls):
        result = None
        if hasattr(cls, "__fields__"):
            result = {}
            fields = getattr(cls, "__fields__")
            for key in fields:
                if hasattr(cls, key):
                    val = getattr(cls, key)
                    result[key] = self.__spread_type(val)
        elif isinstance(cls, dict):
            result = {}
            for item in cls.items():
                k, v = item
                result[k] = self.__spread_type(v)

        elif isinstance(cls, int) or isinstance(cls, float) \
                or isinstance(cls, str) \
                or isinstance(cls, bool):
            result = cls
        elif isinstance(cls, datetime.datetime):
            result = str(cls)
        elif hasattr(cls, "__iter__"):
            result = []
            _iter = cls.__iter__()
            while _iter is not None:
                try:
                    result.append(self.__spread_type(next(_iter)))
                except StopIteration:
                    break
        return result

    def analysis_result(self, result):
        self.__reinit()
        self._source_data = result
        if isinstance(result, tuple):
            self.__gzip_flag = False
        code, plain, headers, data = self.__type_outer(result)
        self.code = code
        if len(plain) > 0:
            self.__plain = plain
        self.__headers.update(headers)
        self.__data = data
        if isinstance(data, types.GeneratorType):
            self.__chunked = True
            self.__sender = data
        elif isinstance(result, FileDeliver) and code == 200:
            self.__path = data
            self.__data = result

    def get_data(self):
        return self._source_data

    def result(self):
        if self.__fd:
            return (self.__path, self.__start_read, self.__read_size)
        elif self.__chunked:
            return self.__sender
        else:
            return self.__data

    def del_cookie(self, name):
        if self.cookies.get(name):
            del self.cookies[name]

    def push_response(self, req: HTTPRequest, single=False):
        tp = str

        if isinstance(self.__data, FileDeliver):
            exter_name = os.path.splitext(self.__path)[-1]
            _range = req.get_header("Range")

            self.__read_size = size = os.path.getsize(self.__path)

            self.__headers["Accept-Ranges"] = 'bytes'
            if _range:
                self.__fd = open(self.__path, "rb")
                _range = _range.split(",")[0].split("=")[1].split('-')
                self.__start_read = start = int(_range[0])
                end = int(_range[1]) if len(_range[1].strip()) > 0 else start + self.__web_cached["range_size"]
                end = size - 1 if end > size else end
                self.__headers["Content-Range"] = "bytes %i-%s/%i" % (
                    start, end, size)
                self.__read_size = size = int(_range[1]) - start if len(_range[1]) > 0 else end - start + 1
                self.__headers["Content-Length"] = str(size)
                self.__range_flag = True
                self.code = 206
                self.__data = b""

            if exter_name in self.__web_cached["extername"]:
                self.__headers["Cache-Control"] = "max-age=%i" % (self.__web_cached["expires"],)
                check_code = 200
                st_mtime = os.stat(self.__path).st_mtime
                lastfiletime = time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.localtime(st_mtime))
                md5 = '\"%s\"' % (hashlib.md5(FileDeliver.read_all(self.__path)).hexdigest(),)

                if req.get_header("If-Modified-Since"):
                    _since = req.get_header("If-Modified-Since").strip()
                    if int(st_mtime) == int(time.mktime(time.strptime(_since, "%a, %d %b %Y %H:%M:%S GMT"))):
                        check_code = 304
                    else:
                        check_code = 200
                elif req.get_header("If-Unmodified-Since"):
                    _since = req.get_header("If-Unmodified-Since").strip()
                    if int(st_mtime) != int(time.mktime(time.strptime(_since, "%a, %d %b %Y %H:%M:%S GMT"))):
                        check_code = 412
                    else:
                        check_code = 304

                if req.get_header("If-None-Match"):
                    _md5 = req.get_header("If-None-Match").strip()
                    if _md5 == md5:
                        check_code = 304
                    else:
                        check_code = 200 if check_code == 200 else check_code
                elif req.get_header("If-Match"):
                    _md5 = req.get_header("If-Match").strip()
                    if _md5 != md5:
                        check_code = 412
                    else:
                        check_code = 304

                if check_code == 304 or check_code == 412:
                    self.__data = b""
                    self.code = check_code if self.code == 200 else self.code
                    self.__chunked = False
                    self.__gzip_flag = False
                    self.__cache_flag = True

                self.__headers["Date"] = self.__headers["Expires"] = time.strftime('%a, %d %b %Y %H:%M:%S GMT',
                                                                                   time.localtime(
                                                                                       time.time() + self.__web_cached[
                                                                                           "expires"]))

                self.__headers["Last-Modified"] = lastfiletime
                self.__headers["Etag"] = md5

            self.__headers["Content-Type"] = HTTPResponse._guess_type(self.__path)

            if not self.__cache_flag and not self.__range_flag:
                if exter_name in self.__gzip_extern_name:
                    self.__data = FileDeliver.read_all(self.__path)
                else:
                    self.__fd = open(self.__path, "rb")
                    self.__data = b""
                    self.__headers["Content-Length"] = size
                    self.__gzip_flag = False
        elif isinstance(self.__data, Exception):
            self.__data = str.join('', traceback.format_exception(self.__data, value=self.__data,
                                                                  tb=self.__data.__traceback__)).encode()

        if len(self.__cookies.keys()):
            self.__headers["Set-Cookie"] = self.cookies

        if self.__chunked and not self.__cache_flag and not self.__range_flag and self.version < 2.0:
            tp = self.__sender.send(None)
            if tp is dict:
                self.__headers["Content-Type"] = "application/json"
            elif tp is str:
                self.__headers["Content-Type"] = "text/plain"
            elif tp is bytes:
                self.__headers["Content-Type"] = "application/octet-stream"
            else:
                self.__headers["Content-Type"] = tp
            self.__data = b""
            self.__gzip_flag = False

        if (self.__gzip_flag or self.__chunked) and (tp is str or tp is dict) and (not self.__range_flag):
            self.__headers["Content-Encoding"] = "gzip"
            if not self.__chunked:
                self.__data = gzip.compress(self.__data if isinstance(self.__data, bytes) else self.__data.encode())

        if self.__fd is None and not self.__cache_flag and not self.__range_flag and not self.__chunked:
            self.__headers["Content-Length"] = len(self.__data)

        if req.method == "HEAD":
            self.__data = b''
            self.__fd = None
            self.__chunked = False
            if self.__chunked:
                del self.__headers["Transfer-Encoding"]
            del self.__headers["Content-Length"]

        if self.version > 1.1:
            return

        header_ = "HTTP/%.1f %i %s\r\n" % (self.version, self.code, self.__plain)

        for item in self.__headers.items():
            key, value = item
            header_ += ("%s: %s\r\n" % (key, value))

        header_ += "\r\n"

        self.control_async(self.write_to_client(header_.encode() + self.__data), single)

        if self.__chunked:
            while True:
                data = self.__sender.send(None)
                if data is None:
                    data = b"0\r\n\r\n"
                    self.control_async(self.write_to_client(data), single)
                    self.__sender.send(None)
                    break
                if tp is dict or tp is str:
                    data = gzip.compress(data.encode())
                    data = hex(len(data))[2:].encode() + b"\r\n" + data + b"\r\n"
                else:
                    data = hex(len(data))[2:].encode() + b"\r\n" + data + b"\r\n"
                asyncio.run_coroutine_threadsafe(self.write_to_client(data), self.__loop)
        elif self.__fd:
            self.control_async(self.sendfile(self.__fd, self.__start_read, self.__read_size), single)

    def control_async(self, fn, single):
        waiter = asyncio.run_coroutine_threadsafe(fn, self.__loop)
        if single:
            return waiter
        else:
            return waiter.result()

    def set_header(self, name, value):
        self.__headers[name] = str(value)

    def del_header(self, name):
        if self.__headers.get(name) is None:
            return
        del self.__headers[name]

    def get_header(self, name, default=None):
        return self.__headers.get(name.lower(), default)

    def headers(self):
        return self.__headers.copy()

    async def sendfile(self, fd, start, count):
        """
        Zero Copy File
        :param fd: File fd
        :param start: start offset
        :param count: read count
        :return:
        """
        try:
            result = await self.__loop.sendfile(self.__writer.transport, fd, offset=start, count=count,
                                                fallback=False)
            fd.close()
            return result
        except PermissionError:
            pass
        except ValueError:
            pass
        except RuntimeError:
            fd.seek(start)
            data = fd.read(count)
            await self.write_to_client(data)
        finally:
            fd.close()

    async def write_to_client(self, data):
        try:
            self.__writer.write(data)
            await self.__writer.drain()
        except ConnectionResetError:
            return



from Scarf.Protocol.HTTP import HTTPRequest, HTTPResponse
from Scarf.Protocol.WebSocket import WebSocketParamter, WebSocket
from Scarf.Context.Timer import TaskParamter
from Scarf.paramter import Paramter
from Scarf.Tip import *

import os, types, traceback


class EventHandler:
    def __init__(self, callback, event, _this):
        self.__callback = callback
        self.__event = event
        self.__source = _this

    def test_event(self, name):
        if name == self.__event:
            return True
        event_ = self.__event.split(".")
        name = name.split(".")
        if len(name) > len(event_):
            return False

        index = -1
        source = 0

        while index >= -len(name):
            if event_[index] == name[index]:
                source += 1
            else:
                break
            index -= 1

        return source

    def set_source(self, _this):
        self.__source = _this

    def __call__(self, *args, **kwargs):
        return self.__callback(self.__source, *args, **kwargs)


class FilterRegister:
    def __init__(self, req, res):
        self.__req_flag_____ = bool(req)
        self.__res_flag_____ = bool(res)
        self.__self_____ = self

    def enter_intercept(self, req: HTTPRequest, res: HTTPResponse):
        if self.__req_flag_____:
            raise RuntimeError("Request Interceptor not implemented")
        else:
            pass

    def outer_filter(self, req: HTTPRequest, res: HTTPResponse):
        if self.__res_flag_____:
            raise RuntimeError("Request Filter not implemented")
        else:
            pass

    def get_reg_opt(self):
        return (self.__req_flag_____, self.__res_flag_____)

    def set_source(self, _this):
        self.__self_____ = _this


class RouterRegister:
    def __init__(self, path, methods, args_source, callback, this, is_file=False):
        self.path = path
        self.methods = methods
        self.is_file = is_file
        self.__args_source = args_source
        self.__this = this
        self.__callback = callback

    def __call__(self, *args, **kwargs):
        return self(self.__this, *args, **kwargs)

    def _get_logger(self, loggers):
        return loggers.get(self.__this.__module__ + "." + self.__this.__class__.__name__)

    def call(self, req: HTTPRequest, res: HTTPResponse, error, sql_use, sqls, extend):
        querys = req.query()
        url_params = req.get_url_params()
        data = None
        ctp = req.get_header("content-type", "")
        if ctp.find("application/x-www-form-urlencoded") > -1:
            data = req.get_data_with_form_urlencoded()
            ctp = ClassSource.FORM_URLENCODE
        elif ctp.find("application/json") > -1:
            data = req.get_data_with_json()
            ctp = ClassSource.JSON
        elif ctp.find("multipart/form-data") > -1:
            data = req.get_data_with_form_data()
            ctp = ClassSource.FORM_DATA
        else:
            ctp = 0
        types_values = self.__callback.__annotations__
        call_params = [self.__this]
        for arg in types_values.items():
            arg_name, tp = arg
            _flag = False
            if isinstance(tp, type) or isinstance(tp, str):
                for _ in extend.items():
                    k, v = _
                    if isinstance(v, type(tp)) or v is tp:
                        _flag = True
                        call_params.append(v)
                        break
                    elif isinstance(tp, str):
                        if k == tp:
                            _flag = True
                            call_params.append(v)
                            break
                if _flag:
                    continue
            if isinstance(tp, HTTPRequest) or tp is HTTPRequest:
                call_params.append(req)
                continue
            elif isinstance(tp, HTTPResponse) or tp is HTTPResponse:
                call_params.append(res)
                continue
            elif tp is bytes:
                call_params.append(req.get_data())
                continue
            elif isinstance(tp, SQLModel.DataBaseConnection):
                if sqls.get(tp.id):
                    if sql_use.get(tp.id):
                        sql_con = sql_use[tp.id]
                    else:
                        sql_con = sql_use[tp.id] = sqls[tp.id].get_con()
                else:
                    sql_con = None
                call_params.append(sql_con)
                continue
            elif tp is SQLModel.DataBaseConnection:
                v = tuple(sqls.items())
                if len(v) > 0:
                    if sql_use.get(v[0][0]):
                        sql_con = sql_use[v[0][0]]
                    else:
                        sql_con = sql_use[v[0][0]] = sqls[v[0][0]].get_con()
                    call_params.append(sql_con)
                else:
                    call_params.append(None)
                continue
            for index, item in enumerate(self.__args_source):
                try:
                    if arg_name in item:
                        if index == 0:
                            call_params.append(tp(querys.get(arg_name, tp())))
                        elif index == 1:
                            if ctp & item[0] == 0:
                                call_params.append(tp())
                                continue
                            if arg_name[0] == "_":
                                _value = data
                            else:
                                _value = data.get(arg_name, tp())
                            if isinstance(tp, Vector):
                                if not isinstance(_value, list):
                                    call_params.append([])
                                    continue
                                try:
                                    call_params.append(self.__arrary_push((data if arg_name[0] == "_" else _value), tp))
                                except:
                                    call_params.append([])
                                continue
                            elif tp is FileDeliver and isinstance(_value, FileDeliver):
                                call_params.append(_value)
                                continue
                            elif tp is dict and isinstance(_value, dict):
                                call_params.append(_value)
                                continue
                            elif isinstance(_value, dict):
                                _new_tp = tp()
                                for _ in _value.items():
                                    key, value = _
                                    if hasattr(_new_tp, key):
                                        if hasattr(_new_tp, "__annotations__"):
                                            cls_type = _new_tp.__annotations__.get(key, type(value))
                                        else:
                                            cls_type = type(value)
                                        try:
                                            value = cls_type(value)
                                        except:
                                            value = tp()
                                        setattr(_new_tp, key, value)
                                call_params.append(_new_tp)
                                continue
                            call_params.append(tp(_value))
                        elif index == 2:
                            call_params.append(tp(req.get_header(arg_name, tp())))
                        elif index == 3:
                            call_params.append(tp(url_params.get(arg_name, tp())))
                except Exception as e:
                    call_params.append(tp())
                    error(e)
                    break
        return self.__callback(*call_params)

    def __arrary_push(self, data, tp):
        params = []
        for item in data:
            if isinstance(tp.type, Vector):
                params.append(self.__arrary_push(item, tp.type))
            elif tp.type is dict:
                params.append(item)
            else:
                _new_type = tp.type()
                for _ in item.items():
                    key, value = _
                    if hasattr(_new_type, key):
                        setattr(_new_type, key, value)
                params.append(_new_type)
        return params


class RequestReslove:
    @staticmethod
    def route(path, methods=(Methods.GET,), args_source=((), (), (), ())):
        def factory(callback=lambda _: _):
            def set_self(self, name) -> RouterRegister:
                router = RouterRegister(path, methods, args_source, callback, self)
                setattr(self, name, router)
                return router

            return set_self

        return factory

    @staticmethod
    def websocket(path, hook, **kwargs):
        def factory(callback=lambda _: _):
            def set_self(self, name) -> WebSocketParamter:
                router = WebSocketParamter(path, hook, self, callback, **kwargs)
                setattr(self, name, router)
                return router

            return set_self

        return factory

    @staticmethod
    def event(*args, **externs):
        def factory(fn=lambda _: _):

            def set_self(self, name) -> EventHandler:
                router = EventHandler(fn, args[0], self)
                setattr(self, name, router)
                return router

            if isinstance(fn, types.FunctionType) or isinstance(fn, types.MethodType):
                return set_self
            else:
                module_rename = externs.get("modules", False)
                if len(args) % 2 == 0 and len(args) > 0:
                    index = 0
                    while index < len(args):
                        event_name = args[index]
                        if module_rename:
                            event_name = fn.__module__ + "." + fn.__name__ + "." + event_name
                            fn_ = getattr(fn, args[index + 1])
                            setattr(fn, args[index + 1], RequestReslove.event(event_name)(fn_))
                        index += 2
                return fn

        return factory

    @staticmethod
    def timer(cycle, times=0, user_arg=None):
        def factory(callback=lambda _: _):
            def set_self(self, name) -> TaskParamter:
                if isinstance(cycle, int):
                    task = TaskParamter(cycle, times, callback, self, user_arg)
                else:
                    raise TypeError("Unknow Type In Timer")
                setattr(self, name, task)
                return task

            return set_self

        return factory

    def __init__(self, routers, error_, config: Paramter, filters, sql, extend, loggers):
        self.__routers = routers
        self.__loggers = loggers
        self.__sql = sql
        self.__enter_intercept = []
        self.__outer_filters = []
        self.__extend = extend
        enter_intercept, outer_filters = filters
        self.__enter_intercept = enter_intercept
        self.__outer_filters = outer_filters
        self.__config = config
        self.__single_reslove = self.__config.get_thread_pool_worker_num().get("single", False)
        self.__cros_config = self.__config.get_cros_params()
        static_path, entry_file = self.__config.get_static_path()
        self.__static_path = static_path
        self.__static_entryfile = entry_file
        self.__error = error_

    def _get_global_sql(self):
        return self.__sql

    def __get_deep_from_method(self, fn):
        result = []
        extends = {}
        for i in fn.__annotations__.items():
            k, v = i
            if isinstance(v, SQLModel.DataBaseConnection):
                sql = self.__sql.get(v.id)
                if sql:
                    result = (k, v.id)
                else:
                    result = (k, None)
            elif v is SQLModel.DataBaseConnection:
                v = tuple(self.__sql.items())
                if len(v) > 0:
                    result = (k, v[0][0])
                else:
                    result = (k, None)
            else:
                for _ in self.__extend.items():
                    _k, _v = _
                    if isinstance(v, str):
                        if _k == v:
                            extends[k] = _v
                            break
                    elif isinstance(_v, type(v)) or v is _v:
                        extends[k] = _v
                        break
        return (result, extends)

    def __close_cons(self, sql_use):
        for item in sql_use.items():
            n, s = item
            self.__sql[n].destory_con(s)

    def router_reslove(self, req: HTTPRequest, res: HTTPResponse):
        sql_use = {}

        router = self._router_match(req)

        req.is_file = isinstance(router, str) or (isinstance(router, RouterRegister) and router.is_file)
        req.is_websocket = isinstance(router, WebSocketParamter)
        req.is_not_found = router is None

        for intercept in self.__enter_intercept:
            sql_params = {}
            result = self.__get_deep_from_method(intercept.enter_intercept)
            if len(result[0]) > 1:
                arg_name, id = result[0]
                if arg_name and id:
                    if sql_use.get(id):
                        sql_params[arg_name] = sql_use[id]
                    else:
                        sql_params[arg_name] = sql_use[id] = self.__sql[id].get_con()
                else:
                    sql_params[arg_name] = None
            result = intercept.enter_intercept(req, res, **sql_params, **result[1])
            if result == 0 or result is None:
                continue
            elif result == 1:
                res.analysis_result(res._source_data)
                res.push_response(req, self.__single_reslove)
                if req.http_version > 1.1:
                    yield None
                self.__close_cons(sql_use)
                yield None
            elif result == 2:
                self.__close_cons(sql_use)
                raise ConnectionRefusedError()

        if router is None:
            res.set_data((404, {"Content-Type": "text/plain"}, "Not Found"))
        elif isinstance(router, str):
            res.set_data(FileDeliver(router))
        elif req.method.upper() == "OPTIONS":
            for _ in self.__cros_config.items():
                key, value = _
                res.set_header(key, value)
            res.set_header("Access-Control-Allow-Methods", str.join(', ', router.methods))
        elif req.method.upper() not in str.join(', ', router.methods):
            res.analysis_result((405, {"Content-Type": "text/plain"}, "Method Not Allowed"))
        elif isinstance(router, RouterRegister):
            try:
                res.set_data(router.call(req, res, self.__error, sql_use, self.__sql, self.__extend))
            except Exception as e:
                logger = router._get_logger(self.__loggers)
                if logger:
                    logger.error(str.join('', traceback.format_exception(e, value=e, tb=e.__traceback__)))
                res.set_data(e)

        elif isinstance(router, WebSocketParamter):
            key = req.get_header("Sec-WebSocket-Key")
            if req.get_header("Connection") and key:
                router = WebSocket(router.path, router.hook, router.hook[0].get_logger(self.__loggers), **router.kwargs,
                                   single=self.__single_reslove)
                result = router(req, key, sql=self.__sql, sql_use=sql_use)
                if isinstance(result, tuple):
                    res.analysis_result(result)
                else:
                    res.analysis_result(router._handshake())
                res.push_response(req, self.__single_reslove)
                yield router
                yield None
            else:
                res.set_data((400, {"Content-Type": "text/plain"}, b"Invaild WebSocket Request"))

        for outer in self.__outer_filters:
            sql_params = {}
            result = self.__get_deep_from_method(outer.outer_filter)
            if len(result[0]) > 1:
                arg_name, id = result[0]
                if arg_name and id:
                    if sql_use.get(id):
                        sql_params[arg_name] = sql_use[id]
                    else:
                        sql_params[arg_name] = sql_use[id] = self.__sql[id].get_con()
                else:
                    sql_params[arg_name] = None
            result = outer.outer_filter(req, res, **sql_params, **result[1])
            if result == 0 or result is None:
                continue
            elif result == 1:
                break
            elif result == 2:
                self.__close_cons(sql_use)
                raise ConnectionRefusedError()

        res.analysis_result(res._source_data)
        res.push_response(req, self.__single_reslove)
        if req.http_version > 1.1:
            yield None
        self.__close_cons(sql_use)
        yield None

    def _router_match(self, req):
        _path = req.path
        path = self.__static_path + _path.replace("\\", "/")
        if os.path.exists(path):
            if os.path.isfile(path):
                return path
            else:
                if os.path.exists(path + "/" + self.__static_entryfile):
                    return path + "/" + self.__static_entryfile

        for item in self.__routers:
            url_params = {}

            _item = [_.strip() for _ in item.path.split("/") if len(_.strip()) > 0]
            path = [_.strip() for _ in _path.split("/") if len(_.strip()) > 0]

            if len(path) > len(_item) and _item[-1] != "*":
                continue
            elif len(path) < len(_item):
                for n in range(0, len(_item) - len(path)):
                    path.append(None)

            for i, _ in enumerate(_item):
                if _[0] == ":":
                    url_params[_item[i][1:]] = path[i]
                elif _ != "*" and path[i] != _:
                    break
                if i == len(_item) - 1:
                    req.set_url_params(url_params)
                    return item
        return None



from Scarf.Tip import FileDeliver
import yaml, os, multiprocessing


class Paramter:
    def __init__(self, configpath, config=None):
        if configpath:
            with open(configpath, "r+", encoding='utf8') as fd:
                self.__paramters = yaml.safe_load(fd.read())
        else:
            self.__paramters = config
        self.__release = self.__GetProperties(False, "server", "release")
        self.__http_port = self.__GetProperties(81, "server", "ports", "http")
        self.__https_port = self.__GetProperties(-1, "server", "ports", "https")
        self.__http2 = self.__GetProperties({"enable": False}, "server", "http2")
        self.__thread_pool_workers = self.__GetProperties(
            {"min": multiprocessing.cpu_count() * 2, "max": multiprocessing.cpu_count() * 2 + 3, "max_work_task": 10},
            "server", "workers")
        self.__server_ssl = {
            "crt": self.__GetProperties(None, "server", "ssl_options", "crt"),
            "key": self.__GetProperties(None, "server", "ssl_options", "key"),
            "pwd": self.__GetProperties('', "server", "ssl_options", "pwd"),
            "ssl_handshake_timeout": self.__GetProperties(15, "server", "ssl_options", "ssl_handshake_timeout")
        }
        self.__gzip = self.__GetProperties(tuple(), "server", "gzip")
        self.__static_options = [self.__GetProperties("", "server", "static", "visitpath").replace("\\", "/"),
                                 self.__GetProperties("index.html", "server", "static", "entryfile")]
        if os.path.exists(self.__static_options[0]):
            self.__static_options[0] = os.path.abspath(self.__static_options[0]).replace("\\", "/")
        self.__sql_options = []
        self.__keep_alive_timeout = (self.__GetProperties(15, "server", "keep_alive", "http_handshake_timeout"),
                                     self.__GetProperties(3, "server", "keep_alive", "http_data_readtimeout"),
                                     self.__GetProperties(3, "server", "keep_alive", "alive_timeout"))
        for item in self.__GetProperties([], "datasource"):
            self.__sql_options.append(item)
        cros_params = ("Access-Control-Allow-Origin", "Access-Control-Allow-Headers", "Access-Control-Max-Age",
                       "Access-Control-Allow-Credentials")
        self.__cros_params = {}
        self.__webcached = self.__GetProperties(
            {"expires": 0, "extername": tuple(), "range_size": 3145728}, "server",
            "webcached")
        self.__Pages = self.__GetProperties([], "server", "static", "map")
        for index, item in enumerate(self.__GetProperties([], "server", "cros")):
            if len(str(item)) > 0:
                self.__cros_params[cros_params[index]] = str(item).lower() if index == 3 else str(item)

    def __GetProperties(self, default_value, *keys):
        _item = self.__paramters
        for index, item in enumerate(keys):
            _item = _item.get(item, default_value) if index == len(keys) - 1 else _item.get(item, {})
        return _item

    def get_running_time_port(self):
        try:
            self.__http_port = int(self.__http_port)
            self.__https_port = int(self.__https_port)
            if self.__http_port > 65535 or self.__https_port > 65535:
                raise Exception("Error Port")
        except Exception as e:
            raise e
        """Get (HTTPPort,HTTPSPort,)"""
        return (self.__http_port, self.__https_port)

    def get_release(self):
        if not isinstance(self.__release, bool):
            raise ValueError("Value Error : release")
        return self.__release

    def get_thread_pool_worker_num(self):
        if not ((self.__thread_pool_workers.get("min") and self.__thread_pool_workers[
            "max"] and self.__thread_pool_workers.get("max_work_task")) or self.__thread_pool_workers.get("single")) :
            raise ValueError("Error TheadPool Options")
        return self.__thread_pool_workers

    def get_ssl_options(self):
        """Get SSLOptions crt,key,password,ssl_hand_shake_time"""
        return self.__server_ssl.copy()

    def get_sql_options(self):
        """Get SQL Options"""
        return self.__sql_options

    def get_static_path(self):
        return self.__static_options

    def get_keep_alive_timeout(self):
        return self.__keep_alive_timeout

    def get_gzip_file_types(self):
        return self.__gzip

    def get_cros_params(self):
        return self.__cros_params

    def get_web_cached(self):
        return self.__webcached

    def get_pages(self):
        return self.__Pages

    def get_http2(self):
        return self.__http2



from Scarf.Protocol.HTTP import HTTPRequest
from Scarf.Tip import SQLModel
from asyncio import StreamWriter

import hashlib, base64, asyncio, socket, traceback


class WebSocketState:
    HANDSHAKE = 0x1,
    READLENGTH = 0x2,
    READDATA = 0x3,
    CLOSE = 0x4,
    INVAILD = 0x5


class WebSocketParamter:
    def __init__(self, path, hook, _this, callback=None, **kwargs):
        self.path = path
        self.hook = hook
        self.__this = _this
        self.methods = ("GET",)
        self.callback = callback
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        return self.callback(self.__this, *args, **kwargs)

    def get_logger(self, loggers):
        return loggers.get(self.__this.__module__ + "." + self.__this.__class__.__name__)

class WebSocket:
    class MessageType:
        CONTINUE = 0x0
        TEXT = 0x1
        BINARY = 0x2
        CLOSE = 0x8
        PING = 0x9
        PONG = 0xA

    def __init__(self, path, hooks, logger, **kwargs):
        self.path = path
        enter_hook, message_hook, close_hook, ping_hook = hooks
        self.__enter_callback = enter_hook
        self.__message_hook = message_hook
        self.__close_hook = close_hook
        self.__ping_hook = ping_hook
        self.__logger = logger
        self._handshake_timeout = kwargs.get("handshake_timeout", 10)
        self._data_timeout = kwargs.get("data_timeout", 10)
        self.state = WebSocketState.HANDSHAKE
        self.__writer: StreamWriter = None
        self.__loop = None
        self.__confim_handshake = False
        self.__run_single = kwargs.get("single", False)
        self.__random = ""
        self.__const_random = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
        self.__sql = {}
        self.__sql_used = {}
        self.user_arg = None

        self.__cache_data = b""
        self.__close_flag = False
        self.__cache_message = []

        self.__fin = 0
        self.__rsvs = [0, 0, 0]
        self.__opcode = 0
        self.__mask_flag = 0
        self.__payload_len = 0
        self.__mask_keys = None
        self.__data = None

    def bind(self, writer, loop, sql):
        self.__writer = writer
        self.__loop = loop
        self.__sql = sql
        for args in self.__cache_message:
            self.__push_data(*args)

    def _handshake(self):
        sha1 = hashlib.sha1()
        sha1.update((self.__random + self.__const_random).encode())
        magic_num = base64.b64encode(sha1.digest()).decode()
        return (101, {"Connection": "Upgrade", "Upgrade": "websocket", "Sec-Websocket-Accept": magic_num}, b"")

    def __get_sql_use(self, fn, sqls):
        for i in fn.__annotations__.items():
            k, v = i
            if isinstance(v, SQLModel.DataBaseConnection):
                sql = sqls.get(v.id)
                if sql:
                    return (k, v.id)
                else:
                    return (k, None)
            elif v is SQLModel.DataBaseConnection:
                v = tuple(sqls.items())
                if len(v) > 0:
                    return (k, v[0][0])
                else:
                    return (k, None)
        return (None, None)

    def _close_sql_cons(self):
        for item in self.__sql_used.items():
            n, s = item
            self.__sql[n].destory_con(s)

    def __close_sql_cons(self, sql_params, cache_params):
        for key in sql_params:
            cache_params[key].destory_con(sql_params[key])

    def __call__(self, data: HTTPRequest, key=None, sql=None, sql_use=None):
        sql_params = {}
        try:
            if self.state == WebSocketState.HANDSHAKE:
                self.__random = key
                arg_name, id = self.__get_sql_use(self.__enter_callback.callback, sql)
                if arg_name:
                    if sql_use.get(id):
                        sql_params[arg_name] = sql_use[id]
                    else:
                        sql_params[arg_name] = sql_use[id] = sql[id].get_con()
                _flag = self.__enter_callback(data, self, **sql_params)
                self.__confim_handshake = True if isinstance(_flag, bool) and _flag else False
                if self.__confim_handshake:
                    self.state = WebSocketState.READLENGTH
                return _flag
            elif self.state == WebSocketState.READLENGTH or self.state == WebSocketState.READDATA:
                self.__reslove_data(data)

        except Exception as e:
            self.__logger.error(str.join('', traceback.format_exception(e, value=e, tb=e.__traceback__)))

    def __call_hook(self, name, *args, **kwargs):
        result = None
        if name == "enter":
            hook = self.__enter_callback
        elif name == "message":
            hook = self.__message_hook
        elif name == "close":
            hook = self.__close_hook
        else:
            hook = self.__ping_hook
        if hook:
            sql_params = {}
            arg_name, id = self.__get_sql_use(hook.callback, self.__sql)
            if arg_name:
                if self.__sql_used.get(id):
                    sql_params[arg_name] = self.__sql_used[id]
                else:
                    sql_params[arg_name] = self.__sql_used[id] = self.__sql[id].get_con()
            kwargs.update(sql_params)
            try:
                result = hook(*args, **kwargs)
            except Exception as e:
                self.__logger.error(str.join('', traceback.format_exception(e, value=e, tb=e.__traceback__)))
        return result

    def __reslove_data(self, data):
        if len(self.__cache_data) > 0:
            data = self.__cache_data + data
            self.__cache_data = b""
        if len(data) < 2:
            self.__cache_data += data
            self.state = WebSocketState.READLENGTH
            return
        else:
            header_size = 2
            if self.state == WebSocketState.READDATA:
                if self.__payload_len < 126:
                    header_size = 6
                elif self.__payload_len <= 65536:
                    header_size = 8
                else:
                    header_size = 14
                if self.__payload_len + header_size > len(data):
                    self.state = WebSocketState.READDATA
                    self.__cache_data = data
                    return
            else:
                self.__fin = (data[0] & 0x80) >> 7
                self.__rsvs[0] = (data[0] & 0x40) >> 6
                self.__rsvs[1] = (data[0] & 0x20) >> 5
                self.__rsvs[2] = (data[0] & 0x10) >> 4
                self.__opcode = (data[0]) & 0xf
                self.__mask_flag = (data[1] & 0x80) >> 7
                if not self.__mask_flag:
                    self.state = WebSocketState.INVAILD
                    return
                payload_len = data[1] & 0x7f
                if payload_len == 126:
                    if len(data) < 4:
                        self.state = WebSocketState.READLENGTH
                        self.__cache_data = data
                        return
                    header_size += 2
                    payload_len = int.from_bytes(data[2:4], byteorder='big', signed=False)
                elif payload_len == 127:
                    if len(data) < 10:
                        self.state = WebSocketState.READLENGTH
                        self.__cache_data = data
                        return
                    header_size += 8
                    payload_len = int.from_bytes(data[2:10], byteorder='big', signed=False)
                self.__payload_len = payload_len
                if len(data) < header_size + 4:
                    self.state = WebSocketState.READLENGTH
                    self.__cache_data = data
                    return
                self.__mask_keys = [int(item) for item in data[header_size:header_size + 4]]
                header_size += 4
                if len(data) < header_size + self.__payload_len:
                    self.state = WebSocketState.READDATA
                    self.__cache_data = data
                    return

            payload_data = bytearray(data[header_size:header_size + self.__payload_len])
            for index, _ in enumerate(payload_data):
                payload_data[index] = self.__mask_keys[index % 4] ^ _
            if self.__opcode == WebSocket.MessageType.TEXT:
                self.__data = payload_data.decode()
            elif self.__opcode == WebSocket.MessageType.BINARY:
                self.__data = bytes(payload_data)
            if self.__opcode == WebSocket.MessageType.CLOSE:
                self.__data = bytes(payload_data)
                code = -1
                reason = ""
                if len(self.__data) > 0:
                    code = int.from_bytes(self.__data[0:2], byteorder='big', signed=False)
                    reason = self.__data[2:].decode()
                self.__close_flag = True
                self.__call_hook("close", self, code, reason)
            elif self.__opcode == WebSocket.MessageType.PONG:
                self.__call_hook("ping", self)
            else:
                self.__call_hook("message", self, self.__data, self.__fin)
            self.state = WebSocketState.READLENGTH
            if len(data[header_size + self.__payload_len:]) > 0:
                self.__reslove_data(data[header_size + self.__payload_len:])

    def test_confirm(self):
        return self.__confim_handshake

    def ping(self):
        self.__push_data(b"", WebSocket.MessageType.PING)

    def send(self, data):
        if isinstance(data, str):
            self.__push_data(data, WebSocket.MessageType.TEXT)
        else:
            self.__push_data(data, WebSocket.MessageType.BINARY)

    def close(self, code, reason):
        self.__push_data(int(code).to_bytes(length=2, byteorder='big', signed=False) + str(reason).encode(),
                         WebSocket.MessageType.CLOSE)

    def __push_data(self, data, msg_type):
        if self.__writer is None:
            self.__cache_message.append((data, msg_type))
            return
        if msg_type == WebSocket.MessageType.TEXT:
            data = data.encode()

        if len(data) < 126:
            _data = bytearray(b"")
            _data.append(0x80 | msg_type)
            _data.append(0x7f & len(data))
            _data = bytes(_data) + data
            self.__sync_write(_data)
        elif len(data) < 0xffff:
            _data = bytearray(b"")
            _data.append(0x80 | msg_type)
            _data.append(126)
            _data = bytes(_data) + len(data).to_bytes(length=2, byteorder='big', signed=False)
            _data += data
            self.__sync_write(_data)
        else:
            _count = len(data) // 131000 + 1 if len(data) % 131000 > 0 else len(data) / 131000
            for _ in range(0, _count):
                _data = bytearray(b"")
                slice_data = data[131000 * _: 131000 * (_ + 1)]
                if _ == 0:
                    _data.append(msg_type)
                elif _ == _count - 1:
                    if len(slice_data) < 0xffff:
                        self.__push_data(slice_data, WebSocket.MessageType.CONTINUE)
                        break
                    _data.append(0x80 | msg_type)
                else:
                    _data.append(WebSocket.MessageType.CONTINUE)
                _data.append(127)
                _data = bytes(_data) + len(slice_data).to_bytes(length=8, byteorder='big', signed=False)
                _data += slice_data
                self.__sync_write(_data)

        if msg_type == WebSocket.MessageType.CLOSE:
            self.state = WebSocketState.CLOSE
            self.__writer.transport.get_extra_info("socket").shutdown(socket.SHUT_WR)

    def __sync_write(self, data):
        waiter = asyncio.run_coroutine_threadsafe(self.__write_data(data), self.__loop)
        if self.__run_single:
            return waiter
        else:
            return waiter.result()

    async def __write_data(self, data):
        self.__writer.write(data)
        await self.__writer.drain()



from Scarf.Protocol.HTTP import HTTPRequest, HTTPResponse
from asyncio import StreamWriter

import asyncio, types, math, random, gzip, sys

sys.setrecursionlimit(0xffff)


class HTTP2ConnectionState:
    OPEN = 0x0
    CONTINUE = 0x1
    RESLOVE_REQUEST = 0x2
    INVAILD = 0x3


class NodeItem:
    def __init__(self, prev, next, value):
        self.prev = prev
        self.next = next
        self.value = value


class Stream:
    def __init__(self, _id, window_size):
        self.state = None
        self.id = _id
        self.cache_data = b""
        self.weight = 0xff
        self.window_size = window_size
        self.frame_size = 0x4000
        self.monopoly = False
        self.http_request: HTTPRequest = None
        self.deep_id = 0
        self.open_stream = True
        self.result_frames = []


class HPACK:
    huffman_table = None
    static_table = [{
        ":authority": ""
    }, {
        ":method": "GET"
    }, {
        ":method": "POST"
    }, {
        ":path": "/"
    }, {
        ":path": "/index.html"
    }, {
        ":scheme": "http"
    }, {
        ":scheme": "https"
    }, {
        ":status": "200"
    }, {
        ":status": "204"
    }, {
        ":status": "206"
    }, {
        ":status": "304"
    }, {
        ":status": "400"
    }, {
        ":status": "404"
    }, {
        ":status": "500"
    }, {
        "accept-charset": ""
    }, {
        "accept-encoding": "gzip, deflate"
    }, {
        "accept-language": ""
    }, {
        "accept-ranges": ""
    }, {
        "accept": ""
    }, {
        "access-control-allow-origin": ""
    }, {
        "age": ""
    }, {
        "allow": ""
    }, {
        "authorization": ""
    }, {
        "cache-control": ""
    }, {
        "content-disposition": ""
    }, {
        "content-encoding": ""
    }, {
        "content-language": ""
    }, {
        "content-length": ""
    }, {
        "content-location": ""
    }, {
        "content-range": ""
    }, {
        "content-type": ""
    }, {
        "cookie": ""
    }, {
        "date": ""
    }, {
        "etag": ""
    }, {
        "expect": ""
    }, {
        "expires": ""
    }, {
        "from": ""
    }, {
        "host": ""
    }, {
        "if-match": ""
    }, {
        "if-modified-since": ""
    }, {
        "if-none-match": ""
    }, {
        "if-range": ""
    }, {
        "if-unmodified-since": ""
    }, {
        "last-modified": ""
    }, {
        "link": ""
    }, {
        "location": ""
    }, {
        "max-forwards": ""
    }, {
        "proxy-authenticate": ""
    }, {
        "proxy-authorization": ""
    }, {
        "range": ""
    }, {
        "referer": ""
    }, {
        "refresh": ""
    }, {
        "retry-after": ""
    }, {
        "server": ""
    }, {
        "set-cookie": ""
    }, {
        "strict-transport-security": ""
    }, {
        "transfer-encoding": ""
    }, {
        "user-agent": ""
    }, {
        "vary": ""
    }, {
        "via": ""
    }, {
        "www-authenticate": ""
    }]
    static_huffman_table = {
        "1111111111000": 0,
        "11111111111111111011000": 1,
        "1111111111111111111111100010": 2,
        "1111111111111111111111100011": 3,
        "1111111111111111111111100100": 4,
        "1111111111111111111111100101": 5,
        "1111111111111111111111100110": 6,
        "1111111111111111111111100111": 7,
        "1111111111111111111111101000": 8,
        "111111111111111111101010": 9,
        "111111111111111111111111111100": 10,
        "1111111111111111111111101001": 11,
        "1111111111111111111111101010": 12,
        "111111111111111111111111111101": 13,
        "1111111111111111111111101011": 14,
        "1111111111111111111111101100": 15,
        "1111111111111111111111101101": 16,
        "1111111111111111111111101110": 17,
        "1111111111111111111111101111": 18,
        "1111111111111111111111110000": 19,
        "1111111111111111111111110001": 20,
        "1111111111111111111111110010": 21,
        "111111111111111111111111111110": 22,
        "1111111111111111111111110011": 23,
        "1111111111111111111111110100": 24,
        "1111111111111111111111110101": 25,
        "1111111111111111111111110110": 26,
        "1111111111111111111111110111": 27,
        "1111111111111111111111111000": 28,
        "1111111111111111111111111001": 29,
        "1111111111111111111111111010": 30,
        "1111111111111111111111111011": 31,
        "010100": 32,
        "1111111000": 33,
        "1111111001": 34,
        "111111111010": 35,
        "1111111111001": 36,
        "010101": 37,
        "11111000": 38,
        "11111111010": 39,
        "1111111010": 40,
        "1111111011": 41,
        "11111001": 42,
        "11111111011": 43,
        "11111010": 44,
        "010110": 45,
        "010111": 46,
        "011000": 47,
        "00000": 48,
        "00001": 49,
        "00010": 50,
        "011001": 51,
        "011010": 52,
        "011011": 53,
        "011100": 54,
        "011101": 55,
        "011110": 56,
        "011111": 57,
        "1011100": 58,
        "11111011": 59,
        "111111111111100": 60,
        "100000": 61,
        "111111111011": 62,
        "1111111100": 63,
        "1111111111010": 64,
        "100001": 65,
        "1011101": 66,
        "1011110": 67,
        "1011111": 68,
        "1100000": 69,
        "1100001": 70,
        "1100010": 71,
        "1100011": 72,
        "1100100": 73,
        "1100101": 74,
        "1100110": 75,
        "1100111": 76,
        "1101000": 77,
        "1101001": 78,
        "1101010": 79,
        "1101011": 80,
        "1101100": 81,
        "1101101": 82,
        "1101110": 83,
        "1101111": 84,
        "1110000": 85,
        "1110001": 86,
        "1110010": 87,
        "11111100": 88,
        "1110011": 89,
        "11111101": 90,
        "1111111111011": 91,
        "1111111111111110000": 92,
        "1111111111100": 93,
        "11111111111100": 94,
        "100010": 95,
        "111111111111101": 96,
        "00011": 97,
        "100011": 98,
        "00100": 99,
        "100100": 100,
        "00101": 101,
        "100101": 102,
        "100110": 103,
        "100111": 104,
        "00110": 105,
        "1110100": 106,
        "1110101": 107,
        "101000": 108,
        "101001": 109,
        "101010": 110,
        "00111": 111,
        "101011": 112,
        "1110110": 113,
        "101100": 114,
        "01000": 115,
        "01001": 116,
        "101101": 117,
        "1110111": 118,
        "1111000": 119,
        "1111001": 120,
        "1111010": 121,
        "1111011": 122,
        "111111111111110": 123,
        "11111111100": 124,
        "11111111111101": 125,
        "1111111111101": 126,
        "1111111111111111111111111100": 127,
        "11111111111111100110": 128,
        "1111111111111111010010": 129,
        "11111111111111100111": 130,
        "11111111111111101000": 131,
        "1111111111111111010011": 132,
        "1111111111111111010100": 133,
        "1111111111111111010101": 134,
        "1011001": 135,
        "1111111111111111010110": 136,
        "11111111111111111011010": 137,
        "11111111111111111011011": 138,
        "11111111111111111011100": 139,
        "11111111111111111011101": 140,
        "11111111111111111011110": 141,
        "111111111111111111101011": 142,
        "11111111111111111011111": 143,
        "111111111111111111101100": 144,
        "111111111111111111101101": 145,
        "1111111111111111010111": 146,
        "11111111111111111100000": 147,
        "111111111111111111101110": 148,
        "11111111111111111100001": 149,
        "11111111111111111100010": 150,
        "11111111111111111100011": 151,
        "11111111111111111100100": 152,
        "111111111111111111100": 153,
        "1111111111111111011000": 154,
        "11111111111111111100101": 155,
        "1111111111111111011001": 156,
        "11111111111111111100110": 157,
        "11111111111111111100111": 158,
        "111111111111111111101111": 159,
        "1111111111111111011010": 160,
        "111111111111111111101": 161,
        "11111111111111111001": 162,
        "1111111111111111011011": 163,
        "1111111111111111011100": 164,
        "11111111111111111101000": 165,
        "11111111111111111101001": 166,
        "111111111111111111110": 167,
        "11111111111111111101010": 168,
        "1111111111111111011101": 169,
        "1111111111111111011110": 170,
        "111111111111111111110000": 171,
        "111111111111111111111": 172,
        "1111111111111111011111": 173,
        "11111111111111111101011": 174,
        "11111111111111111101100": 175,
        "111111111111111100000": 176,
        "111111111111111100001": 177,
        "1111111111111111100000": 178,
        "111111111111111100010": 179,
        "11111111111111111101101": 180,
        "1111111111111111100001": 181,
        "11111111111111111101110": 182,
        "11111111111111111101111": 183,
        "11111111111111111010": 184,
        "1111111111111111100010": 185,
        "1111111111111111100011": 186,
        "1111111111111111100100": 187,
        "11111111111111111110000": 188,
        "1111111111111111100101": 189,
        "1111111111111111100110": 190,
        "11111111111111111110001": 191,
        "11111111111111111111100000": 192,
        "11111111111111111111100001": 193,
        "11111111111111101011": 194,
        "1111111111111110001": 195,
        "1111111111111111100111": 196,
        "11111111111111111110010": 197,
        "1111111111111111101000": 198,
        "1111111111111111111101100": 199,
        "11111111111111111111100010": 200,
        "11111111111111111111100011": 201,
        "11111111111111111111100100": 202,
        "111111111111111111111011110": 203,
        "111111111111111111111011111": 204,
        "11111111111111111111100101": 205,
        "111111111111111111110001": 206,
        "1111111111111111111101101": 207,
        "1111111111111110010": 208,
        "111111111111111100011": 209,
        "11111111111111111111100110": 210,
        "111111111111111111111100000": 211,
        "111111111111111111111100001": 212,
        "11111111111111111111100111": 213,
        "111111111111111111111100010": 214,
        "111111111111111111110010": 215,
        "111111111111111100100": 216,
        "111111111111111100101": 217,
        "11111111111111111111101000": 218,
        "11111111111111111111101001": 219,
        "1111111111111111111111111101": 220,
        "111111111111111111111100011": 221,
        "111111111111111111111100100": 222,
        "111111111111111111111100101": 223,
        "11111111111111101100": 224,
        "111111111111111111110011": 225,
        "11111111111111101101": 226,
        "111111111111111100110": 227,
        "1111111111111111101001": 228,
        "111111111111111100111": 229,
        "111111111111111101000": 230,
        "11111111111111111110011": 231,
        "1111111111111111101010": 232,
        "1111111111111111101011": 233,
        "1111111111111111111101110": 234,
        "1111111111111111111101111": 235,
        "111111111111111111110100": 236,
        "111111111111111111110101": 237,
        "11111111111111111111101010": 238,
        "11111111111111111110100": 239,
        "11111111111111111111101011": 240,
        "111111111111111111111100110": 241,
        "11111111111111111111101100": 242,
        "11111111111111111111101101": 243,
        "111111111111111111111100111": 244,
        "111111111111111111111101000": 245,
        "111111111111111111111101001": 246,
        "111111111111111111111101010": 247,
        "111111111111111111111101011": 248,
        "1111111111111111111111111110": 249,
        "111111111111111111111101100": 250,
        "111111111111111111111101101": 251,
        "111111111111111111111101110": 252,
        "111111111111111111111101111": 253,
        "111111111111111111111110000": 254,
        "11111111111111111111101110": 255,
        "111111111111111111111111111111": 256
    }

    @staticmethod
    def init():
        init_node = NodeItem(None, [None, None], None)
        for item in HPACK.static_huffman_table.items():
            k, v = item
            _init_node = init_node
            for idx, num in enumerate(k):
                num = int(num)
                if num and not _init_node.next[1]:
                    _init_node.next[1] = NodeItem(_init_node, [None, None], None)
                elif num == 0 and not _init_node.next[0]:
                    _init_node.next[0] = NodeItem(_init_node, [None, None], None)
                _init_node = _init_node.next[num]
                if idx == len(k) - 1:
                    _init_node.value = v
        HPACK.huffman_table = init_node

    def __init__(self, size):
        if HPACK.huffman_table is None:
            HPACK.init()
        self.huffman_table = HPACK.huffman_table
        self.static_table = HPACK.static_table.copy()
        self.dynamic_table = []
        self.__dynamic_size = size
        self.dynamic_client_table = []
        self.__dynamic_client_size = size
        self.computed_dynamic(self.dynamic_client_table, size)

    def get_dynamic_size(self):
        return self.__dynamic_size

    def set_dynamic_size(self, val):
        self.__dynamic_size = val
        self.computed_dynamic(self.dynamic_table, val)

    dynamic_size = property(get_dynamic_size, set_dynamic_size)

    def computed_dynamic(self, table, size):
        length = 0
        idx = len(table) - 1
        while idx > -1:
            item = table[idx]
            key = tuple(item.keys())[0]
            if length + len(key) + len(item[key]) + 0x20 > size:
                while idx > -1:
                    table.pop()
                    idx -= 1
                return
            length += len(key) + len(item[key]) + 0x20
            idx -= 1

    def parse_headers(self, data):
        result = {}
        i = 0
        while i < len(data):
            if data[i] >> 5 == 1:
                size = data[i] & 0x1f
                if size == 0x1f:
                    size, i = self.more_repeat(i, data, 0x1f)
                else:
                    self.dynamic_size = size
            elif data[i] >> 7 == 1:
                _index = data[i] & 0x7f
                if _index == 0x7f:
                    _index, i = self.more_repeat(i, data, 0x7f)
                if _index - 1 < len(self.static_table):
                    result.update(self.static_table[_index - 1])
                else:
                    result.update(self.dynamic_table[_index - 1 - len(self.static_table)])
            elif data[i] >> 6 == 1 or data[i] >> 4 == 0 or data[i] >> 4 == 1:
                mode = data[i] >> 6 == 1
                _index = data[i] & (0x3f if mode else 0xf)
                if _index == 0:
                    key_len = data[i + 1] & 0x7f
                    key_huffman = data[i + 1] & 0x80
                    val_len = data[i + 2 + key_len] & 0x7f
                    val_huffman = data[i + 2 + key_len] & 0x80
                    if val_len == 0x7f:
                        val_len, i = self.more_repeat(i + 1, data, 0x7f)
                        i -= 1
                    key = str.join('', [chr(item) for item in (
                        self.huffman_match_str(data[i + 2: i + 2 + key_len]) if key_huffman else data[
                                                                                                 i + 2: i + 2 + key_len])])

                    value = str.join('', [chr(item) for item in (self.huffman_match_str(
                        data[i + 3 + key_len: i + 3 + key_len + val_len]) if val_huffman else data[
                                                                                              i + 3 + key_len: i + 3 + key_len + val_len])])

                    result[key] = value
                    i += (key_len + val_len + 2)
                else:
                    if _index == 0xf:
                        _index, i = self.more_repeat(i, data, 0xf)
                    elif _index == 0x3f:
                        _index, i = self.more_repeat(i, data, 0x3f)
                    if _index - 1 < len(self.static_table):
                        key = tuple(self.static_table[_index - 1].keys())[0]
                    else:
                        key = tuple(self.dynamic_table[_index - 1 - len(self.static_table)].keys())[0]

                    val_len = data[i + 1] & 0x7f
                    huffman = data[i + 1] & 0x80
                    if val_len == 0x7f:
                        val_len, i = self.more_repeat(i + 1, data, 0x7f)
                        i -= 1
                    result[key] = str.join('', [chr(item) for item in (
                        self.huffman_match_str(data[i + 2:i + 2 + val_len], key) if huffman else data[
                                                                                                 i + 2:i + 2 + val_len])])
                    i += (val_len + 1)
                if mode:
                    self.dynamic_table.insert(0, {str(key): str(result[key])})
            i += 1
        self.computed_dynamic(self.dynamic_table, self.dynamic_size)
        return result

    def header_translate(self, headers, table=None, basic=0):
        data = b""
        table = table if table else self.static_table
        for item in headers.items():
            k, v = item
            k = k.lower()
            v = str(v)
            value = 0
            _index = 0
            for i, _ in enumerate(table):
                _k = tuple(_.keys())[0]
                if k == _k and v == _[_k]:
                    value = 0x80
                    i += 1
                    if i < 0x7f:
                        value += (i + basic)
                        data += value.to_bytes(length=1, byteorder='big', signed=False)
                    else:
                        data += self.more_computed(value, (i + basic), 0x7f)
                    break
                elif k == _k and len(_[_k]) == 0:
                    i += 1
                    if i < 0x3f:
                        value = (0x40 + i + basic).to_bytes(length=1, byteorder='big', signed=False)
                    else:
                        value = self.more_computed(0x40, (i + basic), 0x3f)
                    if table is self.static_table:
                        self.dynamic_client_table.append({str(k): str(v)})
                    value += self.huffman_encode(v)
                    data += value
                    break
                if i == len(table) - 1 and table is self.static_table:
                    value = self.header_translate({k: v}, self.dynamic_client_table, len(self.static_table))
                    if len(value) == 0:
                        value = 0x10.to_bytes(length=1, byteorder='big', signed=False)
                        value += self.huffman_encode(k)
                        value += self.huffman_encode(v)
                    data += value
        self.computed_dynamic(self.dynamic_client_table, self.__dynamic_client_size)
        return data

    def more_computed(self, value, idx, basic_num):
        result = b""
        value += basic_num
        result += value.to_bytes(length=1, byteorder='big', signed=False)
        idx -= basic_num
        while idx >= 0x80:
            result += ((idx % 0x80) | 0x80).to_bytes(length=1, byteorder='big', signed=False)
            idx /= 0x80
        result += idx.to_bytes(length=1, byteorder='big', signed=False)
        return result

    def huffman_encode(self, _str):
        result = b''
        data = ""
        for item in _str:
            for huffman in self.static_huffman_table:
                if self.static_huffman_table[huffman] == ord(item):
                    data += huffman
        for _ in range(0, 8 - (len(data) % 8) if len(data) % 8 > 0 else 0):
            data += "1"
        for _idx in range(0, len(data) // 8 + 1 if len(data) % 8 > 0 else len(data) // 8):
            _data = data[_idx * 8:(_idx + 1) * 8]
            _idx = -1
            sum = 0
            while _idx >= -len(_data):
                sum += int(_data[_idx]) * pow(2, _idx * -1 - 1)
                _idx -= 1
            result += sum.to_bytes(length=1, byteorder='big', signed=False)
        result = (0x80 + len(result)).to_bytes(length=1, byteorder='big', signed=False) + result
        return result

    def huffman_match_str(self, data, k=None):
        result_str = []
        cache_node: NodeItem = self.huffman_table
        for item in data:
            _bin_list = []
            while item > 1:
                _bin_list.insert(0, item % 2)
                item = int(item / 2)
            _bin_list.insert(0, item)
            for _ in range(0, 8 - len(_bin_list)):
                _bin_list.insert(0, 0)
            for n, i in enumerate(_bin_list):
                cache_node = cache_node.next[i]
                if cache_node.value:
                    result_str.append(cache_node.value)
                    cache_node = self.huffman_table
        return result_str

    def more_repeat(self, i, data, basic_num):
        cache_data = []
        i += 1
        while data[i] & 0x80 > 0:
            cache_data.insert(0, data[i] & 0x7f)
            i += 1
        _sum = data[i] & 0x7f
        for item in cache_data:
            _sum = _sum * 0x80 + item
        _sum += basic_num
        return (_sum, i)


class HTTP2Context:
    class ErrorCodes:
        NO_ERROR = 0x0
        PROTOCOL_ERROR = 0x1
        INTERNAL_ERROR = 0x2
        FLOW_CONTROL_ERROR = 0x3
        SETTINGS_TIMEOUT = 0x4
        STREAM_CLOSED = 0x5
        FRAME_SIZE_ERROR = 0x6
        REFUSED_STREAM = 0x7
        CANCEL = 0x8
        COMPRESSION_ERROR = 0x9
        CONNECT_ERROR = 0xa
        ENHANCE_YOUR_CALM = 0xb
        INADEQUATE_SECURITY = 0xc
        HTTP_1_1_REQUIRED = 0xd
        error_table = {
            "0": "NO_ERROR",
            "1": "PROTOCOL_ERROR",
            "2": "INTERNAL_ERROR",
            "3": "FLOW_CONTROL_ERROR",
            "4": "SETTINGS_TIMEOUT",
            "5": "STREAM_CLOSED",
            "6": "FRAME_SIZE_ERROR",
            "7": "REFUSED_STREAM",
            "8": "CANCEL",
            "9": "COMPRESSION_ERROR",
            "10": "CONNECT_ERROR",
            "11": "ENHANCE_YOUR_CALM",
            "12": "INADEQUATE_SECURITY",
            "13": "HTTP_1_1_REQUIRED",
        }

    class FrameTypes:
        DATA = 0x0
        HEADERS = 0x1
        PRIORITY = 0x2
        RST_STREAM = 0x3
        SETTINGS = 0x4
        PUSH_PROMISE = 0x5
        PING = 0x6
        GOAWAY = 0x7
        WINDOW_UPDATE = 0x8
        CONTINUATION = 0x9

    settings_params = {
        "SETTINGS_HEADER_TABLE_SIZE": 0x1,
        "SETTINGS_ENABLE_PUSH": 0x2,
        "SETTINGS_MAX_CONCURRENT_STREAMS": 0x3,
        "SETTINGS_INITIAL_WINDOW_SIZE": 0x4,
        "SETTINGS_MAX_FRAME_SIZE": 0x5,
        "SETTINGS_MAX_HEADER_LIST_SIZE": 0x6,
    }

    user_setting_params = {
        "header_size": "SETTINGS_HEADER_TABLE_SIZE",
        "allow_push": "SETTINGS_ENABLE_PUSH",
        "max_stream": "SETTINGS_MAX_CONCURRENT_STREAMS",
        "init_window_size": "SETTINGS_INITIAL_WINDOW_SIZE",
        "max_frame_size": "SETTINGS_MAX_FRAME_SIZE",
        "header_list_size": "SETTINGS_MAX_HEADER_LIST_SIZE"
    }

    def __init__(self, loop, writer: StreamWriter, req: HTTPRequest, request_reslove,
                 project_config, config, error,
                 single: bool):
        self.__single = single
        self.__loop = loop
        self.__writer = writer
        self.__request_reslove = request_reslove
        self.__http_req = req
        self.__error = error
        self.__cache_data = b''
        self.__project_config = project_config
        self.__config = config
        self.__global_settings = [4096, 0, 100, 65535, 0x4000, None]
        self.__activate_frame = {}
        for item in self.__config.get("settings", {}).items():
            k, v = item
            k = HTTP2Context.user_setting_params.get(k)
            if k is None:
                continue
            k = HTTP2Context.settings_params[k]
            # "SETTINGS_HEADER_TABLE_SIZE": 0,
            # "SETTINGS_ENABLE_PUSH": 1,
            # "SETTINGS_MAX_CONCURRENT_STREAMS": 2,
            # "SETTINGS_INITIAL_WINDOW_SIZE": 3,
            # "SETTINGS_MAX_FRAME_SIZE": 4,
            # "SETTINGS_MAX_HEADER_LIST_SIZE": 5,
            self.__global_settings[k - 1] = v
        self.__streams = NodeItem(None, [], Stream(0, self.__global_settings[4]))
        self.__hpack = HPACK(4096)
        self.state = HTTP2ConnectionState.OPEN
        data = req.get_data()
        if data[0: data.find(b"\r\n\r\n")] == b"SM":
            data = data[data.find(b"\r\n\r\n") + 4:]
        else:
            self.state = HTTP2ConnectionState.INVAILD
        # frame format
        self.__length = 0
        self.__type = 0
        self.__flags = 0
        self.__r = 0
        self.__stream_identifier = 0
        self.__extern_data = b""
        self.reslove_data(data)

    def reslove_data(self, data):
        if len(self.__cache_data) > 0:
            data = self.__cache_data + data
            self.__cache_data = b""
        if len(data) < 9:
            self.state = HTTP2ConnectionState.CONTINUE
            self.__cache_data = data
            return
        self.__length = int.from_bytes(data[0:3], byteorder='big', signed=False)
        self.__type = data[3]
        self.__flags = data[4]
        self.__r = data[5] >> 7
        self.__stream_identifier = int.from_bytes(data[5:9], byteorder='big', signed=False) & 0x7fffffff
        stream = self.get_stream(self.__stream_identifier)
        if stream:
            stream = stream.value
            if not stream.open_stream:
                self.close(HTTP2Context.ErrorCodes.PROTOCOL_ERROR,
                           "RST Stream Can Not Reslove Any Message Except PRIORITY")
                return
        elif self.__type != HTTP2Context.FrameTypes.HEADERS:
            self.close(HTTP2Context.ErrorCodes.PROTOCOL_ERROR, "Not Create Stream")
            return
        if len(data) < self.__length:
            self.__cache_data = data
            self.state = HTTP2ConnectionState.CONTINUE
            return
        if self.__length > 0:
            self.__extern_data = data[9: 9 + self.__length]

        if self.__type == HTTP2Context.FrameTypes.SETTINGS:
            self.parse_settings(self.__extern_data, self.__flags, self.__length, stream)
            self.__hpack.dynamic_size = self.__global_settings[0]
        elif self.__type == HTTP2Context.FrameTypes.WINDOW_UPDATE:
            self.parse_window_update(self.__extern_data, self.__length, stream)
        elif self.__type == HTTP2Context.FrameTypes.HEADERS:
            self.parse_headers(self.__extern_data, self.__flags, self.__stream_identifier, self.__length)
        elif self.__type == HTTP2Context.FrameTypes.DATA:
            self.parse_data(self.__extern_data, self.__flags, stream, self.__length)
        elif self.__type == HTTP2Context.FrameTypes.CONTINUATION:
            self.parse_continuation(self.__extern_data, self.__flags, stream)
        elif self.__type == HTTP2Context.FrameTypes.PING:
            self.control_async(self.write_data(self.ping(self.__extern_data, self.__stream_identifier)))
        elif self.__type == HTTP2Context.FrameTypes.GOAWAY:
            self.parse_goaway(self.__extern_data, self.__length)
        elif self.__type == HTTP2Context.FrameTypes.RST_STREAM:
            self.parse_rst_stream(self.__extern_data, stream)
        elif self.__type == HTTP2Context.FrameTypes.PRIORITY:
            self.parse_priority(self.__extern_data, stream)

        data = data[9 + self.__length:]
        if len(data) > 0 and self.state != HTTP2ConnectionState.INVAILD:
            self.reslove_data(data)
        else:
            self.flush()

    def flush(self):
        if self.__activate_frame.get("0"):
            stream = self.__activate_frame["0"]
            self.push_data(stream.result_frames, stream.window_size)
            self.__activate_frame["0"].result_frames.clear()
            del self.__activate_frame["0"]
        if len(self.__activate_frame.keys()):
            weight_pool = {}
            activate_pool = []
            _sum = 0
            for item in self.__activate_frame.values():
                weight_pool[str(item.id)] = []
                if not self.get_node_path(item.id, weight_pool[str(item.id)]):
                    continue
                weight = 0xff + 1
                if len(weight_pool[str(item.id)]) > 1:
                    for index, _ in enumerate(weight_pool[str(item.id)]):
                        sum_weight = sum((_.value.weight for _ in weight_pool[str(item.id)][index - 1].next))
                        weight *= int(_.value.weight / sum_weight) if sum_weight else 1
                        if weight < 5:
                            weight = 5
                            break
                        elif index < len(weight_pool[str(item.id)]) - 1:
                            weight -= 10
                activate_pool.append((weight, item))
                _sum += weight
            result = b""
            while len(activate_pool) > 0:
                verifity_sum = 0
                ran_num = random.randint(0, _sum)
                for idx, item in enumerate(activate_pool):
                    if ran_num < item[0] + verifity_sum:
                        if len(result) + len(item[1].result_frames) >= self.__global_settings[3]:
                            self.control_async(self.write_data(result))
                            result = b""
                        if len(item[1].result_frames) > 0:
                            result += item[1].result_frames.pop(0)
                        if len(item[1].result_frames) == 0:
                            _sum -= item[0]
                            activate_pool.pop(idx)
                            del self.__activate_frame[str(item[1].id)]
                        break
                    verifity_sum += item[0]
            if len(result) > 0:
                self.control_async(self.write_data(result))

    def get_node_path(self, _id, result, start_stream=None):
        start_stream = start_stream if start_stream else self.__streams.next
        for item in start_stream:
            if item.value.id == _id:
                result.append(item)
                return True
            elif len(item.next) > 0:
                if self.get_node_path(_id, result, item.next):
                    result.append(item)
                    return True
        return len(result) > 0

    def parse_priority(self, data, stream):
        if stream.id == 0:
            self.close(HTTP2Context.ErrorCodes.PROTOCOL_ERROR, "PRIORITY Frame StreamIdentifier Must > 0")
            return
        elif len(data) != 5:
            self.close(HTTP2Context.ErrorCodes.FRAME_SIZE_ERROR, "PRIORITY Frame Data Must Have Five Bits")
            return
        node = self.get_stream(stream.id)
        stream: Stream = node.value
        if stream:
            deep_id = int.from_bytes(data[0:4], byteorder='big', signed=False)
            e = deep_id & 0x80000000
            deep_id &= 0x7fffffff
            stream.monopoly = e
            if stream.deep_id != deep_id:
                node.prev.next.remove(node)
                self.get_stream(deep_id).next.append(node)
                stream.deep_id = deep_id
            stream.weight = data[4]

    def parse_goaway(self, data, length):
        # stream_id = int.from_bytes(data[0:3], byteorder='big', signed=False)
        error_code = int.from_bytes(data[3:7], byteorder='big', signed=False)
        plain = data[8: length].decode()
        self.__error(RuntimeError("HTTP2 Error Summary: %s , Error Plain: %s" % (
            HTTP2Context.ErrorCodes.error_table.get(str(error_code), ""), plain)))
        self.__writer.close()

    def parse_continuation(self, data, flags, stream):
        if not stream and (
                stream.state == HTTP2Context.FrameTypes.CONTINUATION or stream.state == HTTP2Context.FrameTypes.HEADERS or stream.state == HTTP2Context.FrameTypes.PUSH_PROMISE):
            self.close(HTTP2Context.ErrorCodes.INTERNAL_ERROR, "HTTP Request Is Not Vaild")
            return
        END_HEADERS = flags & 0x4
        if END_HEADERS and stream.state == HTTP2Context.FrameTypes.HEADERS:
            data = stream.cache_data + data
            try:
                headers = self.__hpack.parse_headers(data)
            except Exception as e:
                self.__error(e)
                self.close(HTTP2Context.ErrorCodes.COMPRESSION_ERROR, str(e))
                return
            self.__create_request(stream, headers)
            stream.state = None
            stream.cache_data = b""
        else:
            stream.cache_data += data

    def parse_data(self, data, flags, stream, length):
        if not stream:
            return
        elif length > self.__global_settings[4]:
            self.close(HTTP2Context.ErrorCodes.PROTOCOL_ERROR, "Frame Size Overflow Settings")
            return
        END_STREAM = flags & 0x1
        PADDED = flags & 0x8
        padded_length = 0
        start = 0

        if PADDED:
            padded_length = data[0]
            start += 1

        _data = data[start: length - padded_length]
        stream.http_request.parse_data(_data)

        if END_STREAM and stream.http_request.has_complate():
            self.__activate_frame[str(stream.id)] = stream
            self.__request_context(stream)
        else:
            self.close(HTTP2Context.ErrorCodes.INTERNAL_ERROR, "HTTP Request Is Not Vaild")

    def parse_headers(self, data, flags, stream_id, length):
        if stream_id > self.__global_settings[2]:
            self.close(HTTP2Context.ErrorCodes.PROTOCOL_ERROR, "Stream Quantity exceeds limit")
            return
        # END_STREAM = flags & 0x1
        END_HEADERS = flags & 0x4
        PADDED = flags & 0x8
        PRIORITY = flags & 0x20
        padded_length = 0
        start = 0

        if PADDED:
            padded_length = data[0]
            start += 1

        if PRIORITY:
            e = data[start] & 128
            deep_stream = int.from_bytes(data[0:4], byteorder='big', signed=False) & 0x7fffffff
            weight = data[start + 4] + 1
            stream: Stream = self.get_stream(stream_id, deep_stream).value
            stream.monopoly = bool(e)
            stream.deep_id = deep_stream
            stream.weight = weight
            stream.frame_size = self.__global_settings[4]
            start += 5
        else:
            stream: Stream = self.get_stream(stream_id, 0).value

        if END_HEADERS:
            header_fragment = stream.cache_data + data[start:length - padded_length]
            headers = self.__hpack.parse_headers(header_fragment)
            self.__create_request(stream, headers)
        else:
            stream.state = HTTP2Context.FrameTypes.HEADERS
            stream.cache_data += data[start:length - padded_length]

    def rst_stream(self, error: int, stream):
        self.control_async(self.write_data(self.http2_outer_(HTTP2Context.FrameTypes.RST_STREAM, 0, stream.id,
                                                             error.to_bytes(length=4, byteorder='big', signed=False))))

    def parse_rst_stream(self, data, stream: Stream):
        if len(data) < 4:
            self.close(HTTP2Context.ErrorCodes.FRAME_SIZE_ERROR, "RST_STREAM Has No Error Code")
            return
        elif stream.id == 0:
            self.close(HTTP2Context.ErrorCodes.PROTOCOL_ERROR, "RST_STREAM Stream Id Must > 0")
            return
        error = int.from_bytes(data[0:4], byteorder='big', signed=False)
        error = HTTP2Context.ErrorCodes.error_table.get(str(error), "")
        self.__error(RuntimeWarning("Stream ID: %i Error, Error summary: %s" % (stream.id, error)))
        stream.result_frames.clear()
        stream.open_stream = False

    def parse_window_update(self, data, length, stream: Stream):
        if length < 4:
            self.close(HTTP2Context.ErrorCodes.FRAME_SIZE_ERROR, "INVAILD Window Size")
            return

        window_size = int.from_bytes(data[0:4], byteorder='big', signed=False) & 0x7fffffff

        if window_size == 0:
            self.close(HTTP2Context.ErrorCodes.PROTOCOL_ERROR, "WINDOW_UPDATE Frame Size Must > 0")
            return
        elif window_size > 0x7fffffff:
            if stream.id == 0:
                self.close(HTTP2Context.ErrorCodes.FLOW_CONTROL_ERROR, "WINDOW UPDATE Exceed Max Size")
            else:
                self.rst_stream(HTTP2Context.ErrorCodes.FLOW_CONTROL_ERROR, stream)
            return

        if self.__stream_identifier == 0:
            if self.__global_settings[0x3] is None:
                self.__global_settings[0x3] = window_size
            else:
                stream.result_frames.append(
                    self.http2_outer_(HTTP2Context.FrameTypes.WINDOW_UPDATE, 0, 0, self.window_update,
                                      self.__global_settings[0x3]))
                self.__activate_frame[str(stream.id)] = stream

        else:
            value: Stream = self.get_stream(self.__stream_identifier).value
            if value:
                value.window_size = window_size

    def parse_settings(self, data, flags, length, stream):
        if flags & 0x1:
            if length > 0:
                self.close(HTTP2Context.ErrorCodes.FRAME_SIZE_ERROR, "Settings ACK Frame Payload Length Must 0")
            return
        for item in range(0, length // 6):
            _data = data[item * 6: (item + 1) * 6]
            identfier = int.from_bytes(_data[0:2], byteorder='big', signed=False) - 1
            value = int.from_bytes(_data[2:6], byteorder='big', signed=False)
            if self.__global_settings[identfier] is None:
                if identfier == 1 and not (value == 0 or value == 1):
                    self.close(HTTP2Context.ErrorCodes.PROTOCOL_ERROR, "SETTINGS_ENABLE_PUSH Must '0' Or '1'")
                    return
                elif identfier == 3 and value > 0x7fffffff:
                    self.close(HTTP2Context.ErrorCodes.FLOW_CONTROL_ERROR, "Flow Control Setting Error")
                    return
                elif identfier == 4 and (value > 0x7fffffff or value < 0xffffff):
                    self.close(HTTP2Context.ErrorCodes.PROTOCOL_ERROR, "Frame Size Setting Error")
                    return
                self.__global_settings[identfier] = value
        stream.result_frames.append(self.http2_outer_(HTTP2Context.FrameTypes.SETTINGS, 0, 0, self.settings))
        stream.result_frames.append(self.http2_outer_(HTTP2Context.FrameTypes.SETTINGS, 1, 0, self.settings, 1))
        self.__activate_frame[str(stream.id)] = stream

    def window_update(self, window_size: int):
        return (window_size & 0x7fffffff).to_bytes(length=4, byteorder='big', signed=False)

    def http2_outer_(self, tp, flags, stream_id, child_fn=None, *args):
        data = tp.to_bytes(length=1, byteorder='big', signed=False)
        data += flags.to_bytes(length=1, byteorder='big', signed=False)
        data += stream_id.to_bytes(length=4, byteorder='big', signed=False)
        if tp == HTTP2Context.FrameTypes.CONTINUATION:
            _data = data
            length = len(_data)
        elif tp == HTTP2Context.FrameTypes.DATA or tp == HTTP2Context.FrameTypes.GOAWAY or tp == HTTP2Context.FrameTypes.RST_STREAM:
            if isinstance(child_fn, int):
                length = child_fn
                _data = b""
            else:
                _data = child_fn
                length = len(_data)
        else:
            _data = child_fn(*args)
            length = len(_data)
        return length.to_bytes(length=3, byteorder='big', signed=False) + data + _data

    def data(self, data, stream: Stream, set_end_stream=True):
        result = b""
        times = math.ceil(len(data) / stream.frame_size)
        for num in range(0, times):
            _data = self.http2_outer_(HTTP2Context.FrameTypes.DATA, int(num == times - 1) and set_end_stream, stream.id,
                                      data[num * stream.frame_size: (num + 1) * stream.frame_size])
            if len(result) + len(_data) > stream.window_size:
                stream.result_frames.append(result)
                result = b""
            result += _data
        if len(result) > 0:
            stream.result_frames.append(result)

    def close(self, error, reason):
        data = self.__stream_identifier.to_bytes(length=4, byteorder='big', signed=False)
        data += error.to_bytes(length=4, byteorder='big', signed=False)
        data += reason.encode()
        self.control_async(self.write_data(self.http2_outer_(HTTP2Context.FrameTypes.GOAWAY, 0, 0, data)))
        self.state = HTTP2ConnectionState.INVAILD
        self.__writer.close()

    def ping(self, data, stream_id, ack=0x1):
        return self.http2_outer_(HTTP2Context.FrameTypes.PING, ack, stream_id, data)

    def headers(self, header_data, pri=False, e=0, deep_stream=0, weight=0):
        data = b""
        if pri:
            data += (deep_stream | (0x80000000 if e else 0)).to_bytes(length=4, byteorder='big', signed=False)
            data += (weight - 1).to_bytes(length=1, byteorder='big', signed=False)
        data += header_data
        return data

    def settings(self, ack=0):
        _data = b""
        if ack == 0:
            for index, item in enumerate(self.__global_settings):
                if item is None:
                    continue
                _data += (index + 1).to_bytes(length=2, byteorder='big', signed=False)
                _data += item.to_bytes(length=4, byteorder='big', signed=False)
        return _data

    def __create_request(self, stream, headers):
        stream.http_request = HTTPRequest(self.__writer, stream.id)
        path = headers.get(":path")
        method = headers.get(":method")
        if path and method:
            request_ = "%s %s HTTP/2.0\r\n" % (method.upper(), path)
            del headers[":path"]
            del headers[":method"]
            for k in headers:
                request_ += "%s: %s\r\n" % (k, headers[k])
            request_ += "\r\n"
            stream.http_request.parse_data(request_.encode())
            if stream.http_request.has_complate():
                self.__activate_frame[str(stream.id)] = stream
                self.__request_context(stream)
        else:
            self.state = HTTP2ConnectionState.INVAILD

    def send_header(self, header_data, stream, end_stream):
        fixed_length = 9
        flags = 0
        pri = False
        if stream.monopoly or stream.weight:
            fixed_length += 5
            flags |= 0x20
            pri = True

        if len(header_data) + fixed_length < stream.window_size:
            flags |= 0x4
            flags |= end_stream
            return [self.http2_outer_(HTTP2Context.FrameTypes.HEADERS, flags, stream.id, self.headers, header_data, pri,
                                      stream.monopoly, stream.deep_id, stream.weight)]
        else:
            result = []
            times = len(header_data) // (stream.window_size - fixed_length) + int(
                bool(len(header_data) % (stream.window_size - fixed_length)))
            for item in range(0, times):
                data = header_data[
                       item * (stream.window_size - fixed_length): (item + 1) * (stream.window_size - fixed_length)]
                if item == 0:
                    data = self.http2_outer_(HTTP2Context.FrameTypes.HEADERS, flags, stream.id, self.headers, data, pri,
                                             stream.monopoly, stream.deep_id, stream.weight)
                elif item == times - 1:
                    flags |= 0x4
                    flags |= end_stream

                data = self.http2_outer_(HTTP2Context.FrameTypes.CONTINUATION, flags, stream.id, data)
                result.append(data)
            return result

    def __request_context(self, stream):
        request = stream.http_request
        response = HTTPResponse(None, None, 2.0, self.__project_config, request.gzip_flag)
        reter = self.__request_reslove.router_reslove(request, response)
        next(reter)
        response_headers = {}
        response_headers[":status"] = str(response.code)
        response_headers.update(response.headers())
        result = response.result()
        tp = None
        if isinstance(result, types.GeneratorType):
            del response_headers["Transfer-Encoding"]
            tp = next(result)
            if tp is dict:
                response_headers["Content-Type"] = "application/json"
            elif tp is str:
                response_headers["Content-Type"] = "text/plain"
            elif tp is bytes:
                del response_headers["Content-Encoding"]
                response_headers["Content-Type"] = "application/octet-stream"
            else:
                response_headers["Content-Type"] = tp
        for source in self.send_header(self.__hpack.header_translate(response_headers), stream,
                                       0 if isinstance(result, types.GeneratorType) or len(result) > 0 else 1):
            stream.result_frames.append(source)
        if isinstance(result, tuple):
            path, start, size = result
            fd = open(path, "rb")
            times = math.ceil(size / stream.frame_size)
            for num in range(0, times):
                _size = size - num * stream.frame_size if num == times - 1 else stream.frame_size
                data = self.http2_outer_(HTTP2Context.FrameTypes.DATA, int(bool(num == times - 1)), stream.id, _size)
                fd.seek(start + num * stream.frame_size)
                data += fd.read(_size)
                stream.result_frames.append(data)
            fd.close()
        elif isinstance(result, types.GeneratorType):
            data = next(result)
            while True:
                end = next(result)
                if tp is dict or tp is str:
                    data = gzip.compress(data.encode())
                self.__activate_frame[str(stream.id)] = stream
                if end is None:
                    self.data(data, stream)
                    self.flush()
                    next(result)
                    break
                self.data(data, stream, False)
                self.flush()
                data = end
        else:
            self.data(result, stream)

        next(reter)
        stream.http_request = None

    def get_stream(self, _id, deep_stream_id=-1, start_stream=None) -> NodeItem:
        """
        Get Stream
        @param _id: Stream Id
        @deep_stream_id: Deep Stream Id - 0: most | >0: deep stream id | <0: No Create(Only Find)
        @weight: Stream Priority
        """
        if _id == 0:
            return self.__streams
        start_stream = start_stream if start_stream else self.__streams.next
        for item in start_stream:
            if item.value.id == _id:
                return item
            elif len(item.next) > 0:
                node_item = self.get_stream(_id, -1, item.next)
                if node_item:
                    return node_item
        node = None
        if deep_stream_id > 0:
            deep_stream = self.get_stream(deep_stream_id, -1, self.__streams.next)
            if deep_stream.value:
                node = NodeItem(deep_stream, [], Stream(_id, self.__global_settings[3]))
                deep_stream.next.append(node)
        elif deep_stream_id == 0:
            node = NodeItem(self.__streams, [], Stream(_id, self.__global_settings[3]))
            self.__streams.next.append(node)
        return node

    def push_data(self, data, window_size):
        result = b""
        for item in data:
            if len(result) + len(item) > window_size:
                self.control_async(self.write_data(result))
                result = b""
            result += item
        if len(result) > 0:
            self.control_async(self.write_data(result))

    def control_async(self, fn):
        waiter = asyncio.run_coroutine_threadsafe(fn, self.__loop)
        if self.__single:
            return waiter
        else:
            return waiter.result()

    async def write_data(self, data):
        try:
            self.__writer.write(data)
            await self.__writer.drain()
        except ConnectionResetError:
            return



from Scarf.paramter import Paramter
from Scarf.Protocol.HTTP2 import HTTP2Context, HTTP2ConnectionState
from Scarf.Protocol.HTTP import HTTPRequest, HTTPResponse, HTTPConnectionState
from Scarf.Protocol.WebSocket import WebSocket, WebSocketState
from Scarf.Context.RequestReslove import RequestReslove

from asyncio import StreamReader, StreamWriter

import asyncio, types


class ClientRequest:
    def __init__(self, context, reader: StreamReader, writer: StreamWriter, loop, config: Paramter,
                 request_reslove: RequestReslove,
                 http2, error: types.FunctionType):
        self.__config = config
        self.__single = config.get_thread_pool_worker_num().get("single", False)
        self.__http2 = http2
        self.__context = context
        self.reader = reader
        self.writer = writer
        self.__error = error
        self.__loop = loop
        self.__request_source = None
        self.__request_reslove = request_reslove
        self.__sql = config.get_sql_options()
        self.handshake_timeout = self.read_data_timeout = self.alive_timeout = 0
        self.__get_config()

    def __get_config(self):
        handshake_timeout, read_data_timeout, alive_timeout = self.__config.get_keep_alive_timeout()
        self.handshake_timeout = handshake_timeout
        self.read_data_timeout = read_data_timeout
        self.alive_timeout = alive_timeout

    def close(self, *args, **kwargs):
        if isinstance(self.__request_source, HTTP2Context):
            self.__request_source.close(*args, **kwargs)

    def read_data(self, data):
        try:
            if self.__request_source is None:
                self.__request_source = request = HTTPRequest(self.reader)
            else:
                request = self.__request_source

            if isinstance(self.__request_source, HTTPRequest):

                request.parse_data(data)

                if not request.has_complate():
                    if request.state == HTTPConnectionState.INVAILD:

                        if request.http_version == 2.0 and self.__http2["enable"]:
                            self.__request_source = HTTP2Context(self.__loop, self.writer, request,
                                                                 self.__request_reslove, self.__config, self.__http2,
                                                                 self.__error,
                                                                 self.__single)
                            if self.__request_source.state != HTTP2ConnectionState.INVAILD:
                                asyncio.run_coroutine_threadsafe(
                                    self.__context.post_read(self, self.reader, self.read_data_timeout),
                                    loop=self.__loop)
                            else:
                                raise ConnectionRefusedError("Close Invaild Connection")
                        return
                    timeout = self.handshake_timeout if request.state == HTTPConnectionState.HEADER else self.read_data_timeout
                    asyncio.run_coroutine_threadsafe(self.__context.post_read(self, self.reader, timeout),
                                                     loop=self.__loop)
                    return

                response = HTTPResponse(self.__loop, self.writer, request.http_version, self.__config,
                                        request.gzip_flag)
                result = next(self.__request_reslove.router_reslove(request, response))
                if isinstance(result, WebSocket):
                    if result.test_confirm():
                        result.bind(self.writer, self.__loop, self.__request_reslove._get_global_sql())
                        self.__request_source = result
                        asyncio.run_coroutine_threadsafe(
                            self.__context.post_read(self, self.reader, result._handshake_timeout),
                            loop=self.__loop)
                    else:
                        self.__request_source = None
                        self.writer.close()
                        raise ConnectionRefusedError("Close Invaild Connection")
                    return
                if request.keep_alive:
                    self.__context.client_context(self.reader, self.writer)
            elif isinstance(self.__request_source, WebSocket):
                self.__request_source(data)
                if self.__request_source.state == WebSocketState.INVAILD or self.__request_source.state == WebSocketState.CLOSE:
                    self.__request_source._close_sql_cons()
                    raise ConnectionRefusedError()
                asyncio.run_coroutine_threadsafe(
                    self.__context.post_read(self, self.reader, self.__request_source._data_timeout),
                    loop=self.__loop)
            elif isinstance(self.__request_source, HTTP2Context):
                self.__request_source.reslove_data(data)
                if self.__request_source.state == HTTP2ConnectionState.INVAILD:
                    raise ConnectionError("HTTP2 Request Invaild")
                else:
                    asyncio.run_coroutine_threadsafe(
                        self.__context.post_read(self, self.reader, self.read_data_timeout),
                        loop=self.__loop)
        except ConnectionError:
            self.writer.close()
        except ConnectionRefusedError:
            self.writer.close()
        except ConnectionResetError:
            self.writer.close()
